# Copyright (C) 2024 Andrea Raffo <andrea.raffo@ibv.uio.no>
#
# SPDX-License-Identifier: MIT

import pathlib
import time
from typing import Dict, List, Sequence, Tuple, Union

import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.sparse as ss
import seaborn as sns
from numpy.typing import NDArray

from . import IO, plot
from .utils import TDA, common, finders, regressions, stripe

be_verbose = True  # TODO consider safe removal of be_verbose


def _log_transform(I: ss.csr_matrix) -> ss.csr_matrix:
    """
    Apply a log-transform to a sparse matrix ignoring (i.e. dropping) NaNs.

    Parameters
    ----------
    I : ss.csr_matrix
        the sparse matrix to be transformed

    Returns
    -------
    ss.csr_matrix
        the log-transformed sparse matrix
    """

    I.data[np.isnan(I.data)] = 0
    I.eliminate_zeros()
    Iproc = I.log1p()
    return Iproc


def _band_extraction(I: ss.csr_matrix, resolution: int, genomic_belt: int) -> Tuple[ss.csr_matrix, ss.csr_matrix]:
    """
    Given a symmetric sparse matrix in CSR format, do the following:

      * Split the input matrix into a upper/lower-triangular matrices
      * Zero (i.e. drop) all values that lie outside the first genomic_belt // resolution diagonals

    Parameters
    ----------
    I : ss.csr_matrix
        the sparse matrix to be processed
    resolution : int
        the genomic resolution of the sparse matrix I
    genomic_belt: int
        the width of the genomic belt to be extracted

    Returns
    -------
    Tuple[ss.csr_matrix, ss.csr_matrix]
        2 elements tuple with the lower-triangular and upper-triangular matrix after band extraction
    """

    assert resolution > 0
    assert genomic_belt > 0

    matrix_belt = genomic_belt // resolution
    LT_I = ss.tril(I, k=0, format="csr") - ss.tril(I, k=-matrix_belt, format="csr")
    UT_I = ss.triu(I, k=0, format="csr") - ss.triu(I, k=matrix_belt, format="csr")
    return LT_I, UT_I


def _scale_Iproc(
    I: ss.csr_matrix, LT_I: ss.csr_matrix, UT_I: ss.csr_matrix
) -> Tuple[ss.csr_matrix, ss.csr_matrix, ss.csr_matrix]:
    """
    Rescale matrices LT_I and UT_I based on the maximum value found in matrix I

    Parameters
    ----------
    I : ss.csr_matrix
        the sparse matrix used to compute the scaling factor
    LT_I : ss.csr_matrix
        the lower-triangular sparse matrix to be rescaled
    UT_I : ss.csr_matrix
        the upper-triangular sparse matrix to be rescaled

    Returns
    -------
    Tuple[ss.csr_matrix, ss.csr_matrix]
        the rescaled lower and upper-triangular matrices
    """

    scaling_factor_Iproc = I.max()
    return tuple(J / scaling_factor_Iproc for J in [I, LT_I, UT_I])  # noqa


def _extract_RoIs(I: ss.csr_matrix, RoI: Dict[str, List[int]]) -> NDArray:
    """
    Extract a region of interest (ROI) from the sparse matrix I

    Parameters
    ----------
    I: ss.csr_matrix
        the sparse matrix to be processed
    RoI: Dict[str, List[int]]
        dictionary with the region of interest in matrix ('matrix') and genomic ('genomic') coordinates

    Returns
    -------
    NDArray
        dense matrix with the interactions for the regions of interest
    """

    rows = cols = slice(RoI["matrix"][0], RoI["matrix"][1])
    I_RoI = I[rows, cols].toarray()
    return I_RoI


def _plot_RoIs(
    I: ss.csr_matrix, Iproc: ss.csr_matrix, RoI: Union[NDArray, None], output_folder: Union[pathlib.Path, None]
) -> Union[NDArray, None]:
    """
    Helper function to plot a region of interest.
    This function does nothing when RoI is None.
    If output_folder is None (but RoI is not), then this function simply extracts the interactions
    that would've been used for plotting, but generates no plots.

    Parameters
    ----------
    I: ss.csr_matrix
        the unprocessed input sparse matrix
    Iproc: ss.csr_matrix
        the processed input sparse matrix
    RoI: Union[NDArray, None]
        the region of interest to be plotted in matrix ('matrix') and genomic ('genomic') coordinates
    output_folder: pathlib.Path
        folder where to save the plots

    Returns
    -------
    Union[NDArray, None]
        the dense matrix used for plotting or None when RoI is None
    """

    # TODO rea1991 Once there is better test coverage, rewrite this as suggested in in #16
    if RoI is not None:
        print("1.4) Extracting a Region of Interest (RoI) for plot purposes...")
        I_RoI = _extract_RoIs(I, RoI)
        Iproc_RoI = _extract_RoIs(Iproc, RoI)

        if output_folder is not None:
            start_pos, end_pos, _, _ = RoI["genomic"]
            # Plots:
            dest = pathlib.Path(output_folder) / f"I_{RoI['genomic'][0]}_{RoI['genomic'][1]}.jpg"
            fig, _, _ = plot.hic_matrix(
                I_RoI,
                (start_pos, end_pos),
                log_scale=False,
            )

            fig.savefig(dest, dpi=256)
            plt.close(fig)

            dest = pathlib.Path(output_folder) / f"Iproc_{RoI['genomic'][0]}_{RoI['genomic'][1]}.jpg"
            fig, _, _ = plot.hic_matrix(
                Iproc_RoI,
                (start_pos, end_pos),
                log_scale=False,
            )
            fig.savefig(dest, dpi=256)
            plt.close(fig)
    else:
        Iproc_RoI = None  # TODO handle this case in _extract_RoIs
    return Iproc_RoI


def _compute_global_pseudodistribution(T: ss.csr_matrix, smooth: bool = True) -> NDArray[float]:
    """
    Given a sparse matrix T, marginalize it, scale the marginal so that maximum is 1, and then smooth it.

    Parameters
    ----------
    T: ss.csr_matrix
        the sparse matrix to be processed
    smooth: bool
        if set to True, smoothing is applied to the pseudo-distribution (default value is True)

    Returns
    -------
    NDArray[np.float64]
        a vector with the re-scaled and smoothed marginals.
    """

    pseudo_dist = np.squeeze(np.asarray(np.sum(T, axis=0)))  # marginalization
    pseudo_dist /= np.max(pseudo_dist)  # scaling
    if smooth:
        pseudo_dist = np.maximum(regressions._compute_wQISA_predictions(pseudo_dist, 11), pseudo_dist)  # smoothing
    return pseudo_dist


def _find_seeds_in_RoI(
    seeds: List[int], left_bound_RoI: int, right_bound_RoI: int
) -> Tuple[NDArray[np.int64], List[int]]:
    # TODO remove
    """
    Select seed coordinates that fall within the given left and right boundaries.

    Parameters
    ----------
    seeds: List[int]
        a list with the seed coordinates
    left_bound_RoI: int
        left bound of the region of interest
    right_bound_RoI: int
        right bound of the region of interest

    Returns
    -------
    Tuple[NDArray[np.int64], List[int]]
        a tuple consisting of:

         * the indices of seed coordinates falling within the given boundaries
         * the coordinates of the selected seeds
    """

    assert left_bound_RoI >= 0
    assert right_bound_RoI >= left_bound_RoI

    # Find sites within the range of interest -- lower-triangular:
    ids_seeds_in_RoI = np.where((left_bound_RoI <= np.array(seeds)) & (np.array(seeds) <= right_bound_RoI))[0]
    seeds_in_RoI = np.array(seeds)[ids_seeds_in_RoI].tolist()

    return ids_seeds_in_RoI, seeds_in_RoI


def _store_results(
    hf: h5py._hl.group.Group,
    pd: NDArray[np.float64],
    min_points: List[int],
    pers_of_min_points: List[float],
    max_points: List[int],
    pers_of_max_points: List[float],
    min_persistence: float,
):
    hf.create_dataset("pseudo-distribution", data=np.array(pd), compression="gzip", compression_opts=4, shuffle=True)
    hf.create_dataset(
        "minima_pts_and_persistence",
        data=np.array([min_points, pers_of_min_points]),
        compression="gzip",
        compression_opts=4,
        shuffle=True,
    )
    hf.create_dataset(
        "maxima_pts_and_persistence",
        data=np.array([max_points, pers_of_max_points]),
        compression="gzip",
        compression_opts=4,
        shuffle=True,
    )
    hf.parent.attrs["min_persistence_used"] = min_persistence


def _check_neighborhood(
    values: NDArray[np.float64], min_value: float = 0.1, neighborhood_size: int = 10, threshold_percentage: float = 0.85
) -> List[int]:
    # TODO rea1991 Change neighborhood size from "matrix" to "genomic" (eg, default of 1 Mb)
    assert 0 <= min_value
    assert 1 <= neighborhood_size <= len(values)
    assert 0 <= threshold_percentage <= 1

    mask = [0] * len(values)
    for i in range(neighborhood_size, len(values) - neighborhood_size):
        neighborhood = values[i - neighborhood_size : i + neighborhood_size + 1]
        ratio_above_min_value = sum(1 for value in neighborhood if value >= min_value) / len(neighborhood)

        if ratio_above_min_value >= threshold_percentage:
            mask[i] = 1
    return mask


def _filter_extrema_by_sparseness(
    ps_mPs: NDArray[int],
    pers_of_ps_mPs: NDArray[float],
    ps_MPs: NDArray[int],
    pers_of_ps_MPs: NDArray[float],
    mask: NDArray[int],
) -> Tuple[NDArray[int], NDArray[float], NDArray[int], NDArray[float]]:
    ps_mPs_2, pers_of_ps_mPs_2, ps_MPs_2, pers_of_ps_MPs_2 = [], [], [], []

    for i in range(len(ps_MPs)):
        if mask[ps_MPs[i]] == 1:
            ps_MPs_2.append(ps_MPs[i])
            pers_of_ps_MPs_2.append(pers_of_ps_MPs[i])

            # Last maximum point is not paired with a minimum point by construction:
            if i < len(ps_MPs) - 1:
                ps_mPs_2.append(ps_mPs[i])
                pers_of_ps_mPs_2.append(pers_of_ps_mPs[i])

    # If last maximum point was discarded, we need to remove the minimum point before it:
    if len(ps_mPs_2) == len(ps_MPs_2) and len(ps_mPs_2) > 0:
        ps_mPs_2.pop()
        pers_of_ps_mPs_2.pop()

    return np.array(ps_mPs_2), np.array(pers_of_ps_mPs_2), np.array(ps_MPs_2), np.array(pers_of_ps_MPs_2)


def step_1(I, genomic_belt, resolution, RoI=None, output_folder=None):
    print("1.1) Log-transformation...")
    Iproc = _log_transform(I)

    print("1.2) Focusing on a neighborhood of the main diagonal...")
    LT_Iproc, UT_Iproc = _band_extraction(Iproc, resolution, genomic_belt)

    print("1.3) Projection onto [0, 1]...")
    Iproc, LT_Iproc, UT_Iproc = _scale_Iproc(Iproc, LT_Iproc, UT_Iproc)

    Iproc_RoI = _plot_RoIs(I, Iproc, RoI, output_folder)

    return LT_Iproc, UT_Iproc, Iproc_RoI


def step_2(chrom: str, L, U, resolution, min_persistence, Iproc_RoI=None, RoI=None, output_folder=None) -> IO.Result:
    print("2.1) Global 1D pseudo-distributions...")
    LT_pd = _compute_global_pseudodistribution(L, smooth=True)
    UT_pd = _compute_global_pseudodistribution(U, smooth=True)

    result = IO.Result(chrom)
    if RoI is not None:
        result.set_roi(RoI)
    result.set_min_persistence(min_persistence)

    result.set("pseudodistribution", LT_pd, "LT")
    result.set("pseudodistribution", UT_pd, "UT")

    print("2.2) Detection of persistent maxima and corresponding minima for lower- and upper-triangular matrices...")

    print("2.2.0) All maxima and their persistence")
    # NOTATION: mPs = minimum points, MPs = maximum Points, ps = persistence-sorted
    # NB: MPs are the actual sites of interest, i.e., the sites hosting linear patterns

    # All local minimum and maximum points:
    all_LT_ps_mPs, all_pers_of_LT_ps_mPs, all_LT_ps_MPs, all_pers_of_LT_ps_MPs = TDA.TDA(LT_pd, min_persistence=0)
    all_UT_ps_mPs, all_pers_of_UT_ps_mPs, all_UT_ps_MPs, all_pers_of_UT_ps_MPs = TDA.TDA(UT_pd, min_persistence=0)

    result.set("all_minimum_points", all_LT_ps_mPs, "LT")
    result.set("all_maximum_points", all_LT_ps_MPs, "LT")
    result.set("persistence_of_all_minimum_points", all_pers_of_LT_ps_mPs, "LT")
    result.set("persistence_of_all_maximum_points", all_pers_of_LT_ps_MPs, "LT")
    result.set("all_minimum_points", all_UT_ps_mPs, "UT")
    result.set("all_maximum_points", all_UT_ps_MPs, "UT")
    result.set("persistence_of_all_minimum_points", all_pers_of_UT_ps_mPs, "UT")
    result.set("persistence_of_all_maximum_points", all_pers_of_UT_ps_MPs, "UT")

    print("2.2.1) Lower triangular part")
    LT_ps_mPs, pers_of_LT_ps_mPs, LT_ps_MPs, pers_of_LT_ps_MPs = TDA.TDA(LT_pd, min_persistence=min_persistence)

    print("2.2.2) Upper triangular part")
    # Here, LT_ps_mPs means that the lower-triangular minimum points are sorted w.r.t. persistence
    # (NOTATION: ps = persistence-sorted)
    UT_ps_mPs, pers_of_UT_ps_mPs, UT_ps_MPs, pers_of_UT_ps_MPs = TDA.TDA(UT_pd, min_persistence=min_persistence)
    # NB: Maxima are sorted w.r.t. their persistence... and this sorting is applied to minima too,
    # so that each maximum is still paired to its minimum.

    # Maximum and minimum points sorted w.r.t. coordinates (NOTATION: cs = coordinate-sorted):
    LT_mPs, LT_pers_of_mPs = common.sort_based_on_arg0(LT_ps_mPs, pers_of_LT_ps_mPs)
    LT_MPs, LT_pers_of_MPs = common.sort_based_on_arg0(LT_ps_MPs, pers_of_LT_ps_MPs)
    UT_mPs, UT_pers_of_mPs = common.sort_based_on_arg0(UT_ps_mPs, pers_of_UT_ps_mPs)
    UT_MPs, UT_pers_of_MPs = common.sort_based_on_arg0(UT_ps_MPs, pers_of_UT_ps_MPs)

    print("2.2.3) Filter out seeds in sparse regions")
    LT_mask = _check_neighborhood(_compute_global_pseudodistribution(L, smooth=False))
    UT_mask = _check_neighborhood(_compute_global_pseudodistribution(U, smooth=False))
    x = _filter_extrema_by_sparseness(LT_mPs, LT_pers_of_mPs, LT_MPs, LT_pers_of_MPs, LT_mask)
    LT_mPs, LT_pers_of_mPs, LT_MPs, LT_pers_of_MPs = x
    x = _filter_extrema_by_sparseness(UT_mPs, UT_pers_of_mPs, UT_MPs, UT_pers_of_MPs, UT_mask)
    UT_mPs, UT_pers_of_mPs, UT_MPs, UT_pers_of_MPs = x
    if len(LT_MPs) < len(LT_ps_MPs):
        print(f"Number of lower-triangular seed sites is reduced from {len(LT_ps_MPs)} to {len(LT_MPs)}")
    if len(UT_MPs) < len(UT_ps_MPs):
        print(f"Number of upper-triangular seed sites is reduced from {len(UT_ps_MPs)} to {len(UT_MPs)}")
    if len(LT_MPs) == len(LT_ps_MPs) and len(UT_MPs) == len(UT_ps_MPs):
        print("No change in number of seed sites")

    result.set("persistent_minimum_points", LT_mPs, "LT")
    result.set("persistent_maximum_points", LT_MPs, "LT")
    result.set("persistence_of_minimum_points", LT_pers_of_mPs, "LT")
    result.set("persistence_of_maximum_points", LT_pers_of_MPs, "LT")

    result.set("persistent_minimum_points", UT_mPs, "UT")
    result.set("persistent_maximum_points", UT_MPs, "UT")
    result.set("persistence_of_minimum_points", UT_pers_of_mPs, "UT")
    result.set("persistence_of_maximum_points", UT_pers_of_MPs, "UT")

    # If no candidates are found in the lower- or upper-triangular maps, exit:
    if len(LT_MPs) == 0 or len(UT_MPs) == 0:
        return result

    print("2.3) Storing into a list of Stripe objects...")
    stripes = [
        stripe.Stripe(seed=LT_MP, top_pers=LT_pers_of_MP, where="lower_triangular")
        for LT_MP, LT_pers_of_MP in zip(LT_MPs, LT_pers_of_MPs)
    ]
    result.set("stripes", stripes, "LT")

    stripes = [
        stripe.Stripe(seed=UT_MP, top_pers=UT_pers_of_MP, where="upper_triangular")
        for UT_MP, UT_pers_of_MP in zip(UT_MPs, UT_pers_of_MPs)
    ]
    result.set("stripes", stripes, "UT")

    if result.roi is not None:
        print("2.4) Finding sites inside the region selected above...")
        ids_LT_MPs_in_RoI, LT_MPs_in_RoI = _find_seeds_in_RoI(LT_MPs, result.roi["matrix"][0], result.roi["matrix"][1])
        ids_UT_MPs_in_RoI, UT_MPs_in_RoI = _find_seeds_in_RoI(UT_MPs, result.roi["matrix"][0], result.roi["matrix"][1])

        print("2.5) Plotting pseudo-distributions and sites for the region selected above...")
        if output_folder is not None:

            # Plot pseudo-distributions:
            IO.pseudodistrib(
                LT_pd, RoI["genomic"][0:2], resolution, output_folder=output_folder, file_name=f"LT_pseudo-distrib.jpg"
            )
            IO.pseudodistrib(
                UT_pd, RoI["genomic"][0:2], resolution, output_folder=output_folder, file_name=f"UT-pseudo-distrib.jpg"
            )

            # Plot pseudo-distributions and persistent maxima:
            IO.pseudodistrib(
                LT_pd,
                RoI["genomic"][0:2],
                resolution,
                coords2scatter=[LT_MPs_in_RoI],
                colors=["blue"],
                output_folder=output_folder,
                title=None,
                file_name=f"LT_pseudo-distrib_and_pers-maxima.jpg",
                display=False,
            )
            IO.pseudodistrib(
                UT_pd,
                RoI["genomic"][0:2],
                resolution,
                coords2scatter=[UT_MPs_in_RoI],
                colors=["blue"],
                output_folder=output_folder,
                title=None,
                file_name=f"UT_pseudo-distrib_and_pers-maxima.jpg",
                display=False,
            )

            # Plot the region of interest of Iproc with over-imposed vertical lines for seeds:
            if Iproc_RoI is not None:
                IO.HiC_and_sites(
                    Iproc_RoI,
                    LT_MPs_in_RoI,
                    RoI["genomic"],
                    resolution,
                    where="lower",
                    plot_in_bp=True,
                    output_folder=output_folder,
                    display=False,
                    file_name=f"LT_seeds.jpg",
                    title=None,
                )
                IO.HiC_and_sites(
                    Iproc_RoI,
                    UT_MPs_in_RoI,
                    RoI["genomic"],
                    resolution,
                    where="upper",
                    plot_in_bp=True,
                    output_folder=output_folder,
                    display=False,
                    file_name=f"UT_seeds.jpg",
                    title=None,
                )

    return result


def step_3(
    result: IO.Result,
    L,
    U,
    resolution,
    genomic_belt,
    max_width,
    constrain_height,
    loc_pers_min,
    loc_trend_min,
    Iproc_RoI=None,
    RoI=None,
    output_folder=None,
    map=map,
) -> IO.Result:
    if result.empty:
        print("3) No candidates found by step 2. Returning immediately!")
        return result

    # Retrieve data:
    LT_mPs = result.get("persistent_minimum_points", "LT")
    UT_mPs = result.get("persistent_minimum_points", "UT")
    LT_MPs = result.get("persistent_maximum_points", "LT")
    UT_MPs = result.get("persistent_maximum_points", "UT")
    LT_pseudo_distrib = result.get("pseudodistribution", "LT")
    UT_pseudo_distrib = result.get("pseudodistribution", "UT")

    start_time = time.time()

    print("3.1) Width estimation")
    print("3.1.1) Estimating widths (equiv. HIoIs, where HIoI stands for Horizontal Interval of Interest)...")

    # Complement mPs with:
    # the global minimum (if any) that is to the left of the leftmost persistent maximum
    # AND
    # the global minimum (if any) that is to the right of the rightmost persistent maximum
    LT_L_nb = np.arange(0, LT_MPs[0])
    LT_R_nb = np.arange(LT_MPs[-1], L.shape[0])
    UT_L_nb = np.arange(0, UT_MPs[0])
    UT_R_nb = np.arange(UT_MPs[-1], U.shape[0])
    LT_L_mP = np.argmin(LT_pseudo_distrib[LT_L_nb]) if len(LT_L_nb) > 0 else -1
    LT_R_mP = LT_MPs[-1] + np.argmin(LT_pseudo_distrib[LT_R_nb]) if len(LT_R_nb) > 0 else -1
    UT_L_mP = np.argmin(UT_pseudo_distrib[UT_L_nb]) if len(UT_L_nb) > 0 else -1
    UT_R_mP = UT_MPs[-1] + np.argmin(UT_pseudo_distrib[UT_R_nb]) if len(UT_R_nb) > 0 else -1

    LT_bounded_mPs = [(max(LT_L_mP, 0),), (max(LT_R_mP, L.shape[0]),)]
    UT_bounded_mPs = [(max(UT_L_mP, 0),), (max(UT_R_mP, U.shape[0]),)]
    # We need to check that the list of minimum points are not empty, otherwise np.concatenate will create an array with dtype=float
    if len(LT_mPs) != 0:
        LT_bounded_mPs.insert(1, LT_mPs)
    if len(UT_mPs) != 0:
        UT_bounded_mPs.insert(1, UT_mPs)

    LT_bounded_mPs = np.concatenate(LT_bounded_mPs, dtype=int)
    UT_bounded_mPs = np.concatenate(UT_bounded_mPs, dtype=int)

    # List of pairs (pair = left and right boundaries):
    # Choose the variable criterion between max_ascent and max_perc_descent
    # ---> When variable criterion is set to max_ascent, set the variable max_ascent
    # ---> When variable criterion is set to max_perc_descent, set the variable max_perc_descent
    LT_HIoIs = finders.find_HIoIs(
        LT_pseudo_distrib, LT_MPs, LT_bounded_mPs, int(max_width / (2 * resolution)) + 1, map=map
    )
    UT_HIoIs = finders.find_HIoIs(
        UT_pseudo_distrib, UT_MPs, UT_bounded_mPs, int(max_width / (2 * resolution)) + 1, map=map
    )

    # List of left or right boundaries:
    LT_L_bounds, LT_R_bounds = map(list, zip(*LT_HIoIs))
    UT_L_bounds, UT_R_bounds = map(list, zip(*UT_HIoIs))

    print("3.1.2) Updating list of Stripe objects with HIoIs...")
    stripes = result.get("stripes", "LT")
    for num_cand_stripe, (LT_L_bound, LT_R_bound) in enumerate(zip(LT_L_bounds, LT_R_bounds)):
        stripes[num_cand_stripe].set_horizontal_bounds(LT_L_bound, LT_R_bound)

    stripes = result.get("stripes", "UT")
    for num_cand_stripe, (UT_L_bound, UT_R_bound) in enumerate(zip(UT_L_bounds, UT_R_bounds)):
        stripes[num_cand_stripe].set_horizontal_bounds(UT_L_bound, UT_R_bound)

    if all([param is not None for param in [RoI, output_folder]]):

        print("3.1.3) Plots")
        # 3.1.3.1 "Finding HIoIs inside the region (RoI) selected above..."

        ids_LT_MPs_in_RoI, LT_MPs_in_RoI = _find_seeds_in_RoI(LT_MPs, result.roi["matrix"][0], result.roi["matrix"][1])
        ids_UT_MPs_in_RoI, UT_MPs_in_RoI = _find_seeds_in_RoI(UT_MPs, result.roi["matrix"][0], result.roi["matrix"][1])

        # Left and right boundaries in RoI:
        LT_L_bounds_in_RoI = np.array(LT_L_bounds)[ids_LT_MPs_in_RoI].tolist()
        LT_R_bounds_in_RoI = np.array(LT_R_bounds)[ids_LT_MPs_in_RoI].tolist()
        UT_L_bounds_in_RoI = np.array(UT_L_bounds)[ids_UT_MPs_in_RoI].tolist()
        UT_R_bounds_in_RoI = np.array(UT_R_bounds)[ids_UT_MPs_in_RoI].tolist()

        # 3.1.3.2 "Plotting pseudo-distributions and sites for the region selected above..."
        IoIs = [RoI["genomic"][0:2]] + [
            [LT_L_bound_in_RoI * resolution, (LT_R_bound_in_RoI + 1) * resolution]
            for (LT_L_bound_in_RoI, LT_R_bound_in_RoI) in zip(LT_L_bounds_in_RoI, LT_R_bounds_in_RoI)
        ]
        IO.pseudodistrib_and_HIoIs(
            LT_pseudo_distrib,
            IoIs,
            resolution,
            colors=["red"] + ["blue"] * len(LT_L_bounds_in_RoI),
            title=None,
            output_folder=output_folder,
            file_name=f"LT_pseudo-distrib_and_h-doms.jpg",
            display=False,
        )
        IoIs = [RoI["genomic"][0:2]] + [
            [UT_L_bound_in_RoI * resolution, (UT_R_bound_in_RoI + 1) * resolution]
            for (UT_L_bound_in_RoI, UT_R_bound_in_RoI) in zip(UT_L_bounds_in_RoI, UT_R_bounds_in_RoI)
        ]
        IO.pseudodistrib_and_HIoIs(
            UT_pseudo_distrib,
            IoIs,
            resolution,
            colors=["red"] + ["blue"] * len(UT_L_bounds_in_RoI),
            title=None,
            output_folder=output_folder,
            file_name=f"UT_pseudo-distrib_and_h-doms.jpg",
            display=False,
        )

        if Iproc_RoI is not None:

            # Projecting left and right boundaries onto the sub-intervals:
            LT_bounds_in_RoI_proj = [
                [max(0, LT_L_bound - RoI["matrix"][0]), min(LT_R_bound - RoI["matrix"][0], Iproc_RoI.shape[0] - 1)]
                for (LT_L_bound, LT_R_bound) in zip(LT_L_bounds_in_RoI, LT_R_bounds_in_RoI)
            ]
            UT_bounds_in_RoI_proj = [
                [max(0, UT_L_bound - RoI["matrix"][0]), min(UT_R_bound - RoI["matrix"][0], Iproc_RoI.shape[0] - 1)]
                for (UT_L_bound, UT_R_bound) in zip(UT_L_bounds_in_RoI, UT_R_bounds_in_RoI)
            ]
            LT_bounds_in_RoI_proj_gen_coord = [[a[0] * resolution, a[1] * resolution] for a in LT_bounds_in_RoI_proj]
            UT_bounds_in_RoI_proj_gen_coord = [[a[0] * resolution, a[1] * resolution] for a in UT_bounds_in_RoI_proj]

            # Slices, i.e., intervals determined by a pair of left & right boundaries:
            LT_slices2keep_proj = [
                list(range(LT_bound_in_RoI_proj[0], LT_bound_in_RoI_proj[1] + 1))
                for LT_bound_in_RoI_proj in LT_bounds_in_RoI_proj
            ]
            UT_slices2keep_proj = [
                list(range(UT_bound_in_RoI_proj[0], UT_bound_in_RoI_proj[1] + 1))
                for UT_bound_in_RoI_proj in UT_bounds_in_RoI_proj
            ]

            # 3.1.3.3 "Plotting RoI restricted to HIoIs..."
            # Setting rows/columns not included in proj_LT_ids_2_keep to zero:

            Iproc0_RoI_LT_sliced = np.triu(Iproc_RoI)
            Iproc0_RoI_UT_sliced = np.tril(Iproc_RoI)
            for num_slice, LT_slice2keep_proj in enumerate(LT_slices2keep_proj):

                Iproc0_RoI_LT_cur_sliced = np.triu(Iproc_RoI)
                for idx2keep in LT_slice2keep_proj:
                    Iproc0_RoI_LT_cur_sliced[idx2keep:, idx2keep] = Iproc_RoI[idx2keep:, idx2keep]
                    Iproc0_RoI_LT_sliced[idx2keep:, idx2keep] = Iproc_RoI[idx2keep:, idx2keep]

            IO.HiC_and_HIoIs(
                Iproc0_RoI_LT_sliced,
                LT_bounds_in_RoI_proj_gen_coord,
                RoI["genomic"],
                resolution,
                title=None,
                output_folder=output_folder,
                plot_in_bp=True,
                where="lower",
                file_name=f"LT_all_h-doms.jpg",
                display=False,
            )

            for num_slice, UT_slice2keep_proj in enumerate(UT_slices2keep_proj):

                Iproc0_RoI_UT_cur_sliced = np.tril(Iproc_RoI)
                for idx2keep in UT_slice2keep_proj:
                    Iproc0_RoI_UT_cur_sliced[: idx2keep + 1, idx2keep] = Iproc_RoI[: idx2keep + 1, idx2keep]
                    Iproc0_RoI_UT_sliced[: idx2keep + 1, idx2keep] = Iproc_RoI[: idx2keep + 1, idx2keep]

            IO.HiC_and_HIoIs(
                Iproc0_RoI_UT_sliced,
                UT_bounds_in_RoI_proj_gen_coord,
                RoI["genomic"],
                resolution,
                title=None,
                output_folder=output_folder,
                plot_in_bp=True,
                where="upper",
                file_name=f"UT_all_h-doms.jpg",
                display=False,
            )

    print(f"Execution time: {time.time() - start_time} seconds ---")

    print("3.2) Height estimation")

    start_time = time.time()

    print("3.2.1) Estimating heights (equiv. VIoIs, where VIoI stands for Vertical Interval of Interest)...")
    if be_verbose and all([param is not None for param in [RoI, output_folder]]):
        LT_VIoIs, LT_peaks_ids = finders.find_VIoIs(
            L,
            LT_MPs,
            LT_HIoIs,
            VIoIs2plot=ids_LT_MPs_in_RoI,
            max_height=int(genomic_belt / resolution),
            threshold_cut=loc_trend_min,
            min_persistence=loc_pers_min,
            where="lower",
            output_folder=f"{output_folder}local_pseudodistributions/",
            map=map,
        )
        UT_VIoIs, UT_peaks_ids = finders.find_VIoIs(
            U,
            UT_MPs,
            UT_HIoIs,
            VIoIs2plot=ids_UT_MPs_in_RoI,
            max_height=int(genomic_belt / resolution),
            threshold_cut=loc_trend_min,
            min_persistence=loc_pers_min,
            where="upper",
            output_folder=f"{output_folder}local_pseudodistributions/",
            map=map,
        )
    else:
        LT_VIoIs, LT_peaks_ids = finders.find_VIoIs(
            L,
            LT_MPs,
            LT_HIoIs,
            VIoIs2plot=None,
            max_height=int(genomic_belt / resolution),
            threshold_cut=loc_trend_min,
            min_persistence=loc_pers_min,
            where="lower",
            output_folder=None,
            map=map,
        )
        UT_VIoIs, UT_peaks_ids = finders.find_VIoIs(
            U,
            UT_MPs,
            UT_HIoIs,
            VIoIs2plot=None,
            max_height=int(genomic_belt / resolution),
            threshold_cut=loc_trend_min,
            min_persistence=loc_pers_min,
            where="upper",
            output_folder=None,
            map=map,
        )

    # List of left or right boundaries:
    LT_U_bounds, LT_D_bounds = map(list, zip(*LT_VIoIs))
    UT_U_bounds, UT_D_bounds = map(list, zip(*UT_VIoIs))

    print("3.2.2) Updating list of Stripe objects with VIoIs...")
    lt_stripes = result.get("stripes", "LT")
    for num_cand_stripe, (LT_U_bound, LT_D_bound) in enumerate(zip(LT_U_bounds, LT_D_bounds)):
        lt_stripes[num_cand_stripe].set_vertical_bounds(LT_U_bound, LT_D_bound)
    ut_stripes = result.get("stripes", "UT")
    for num_cand_stripe, (UT_U_bound, UT_D_bound) in enumerate(zip(UT_U_bounds, UT_D_bounds)):
        ut_stripes[num_cand_stripe].set_vertical_bounds(UT_U_bound, UT_D_bound)

    print(f"Execution time: {time.time() - start_time} seconds ---")

    if RoI is not None:

        print("3.3) Finding HIoIs and VIoIs inside the region (RoI) selected above...")

        # Restricting to the RoI:
        LT_HIoIs_in_RoI = np.array(LT_HIoIs)[ids_LT_MPs_in_RoI].tolist()
        UT_HIoIs_in_RoI = np.array(UT_HIoIs)[ids_UT_MPs_in_RoI].tolist()
        LT_VIoIs_in_RoI = np.array(LT_VIoIs)[ids_LT_MPs_in_RoI].tolist()
        UT_VIoIs_in_RoI = np.array(UT_VIoIs)[ids_UT_MPs_in_RoI].tolist()
        LT_HIoIs_in_RoI_proj = [
            [LT_HIoI_RoI[0] - RoI["matrix"][0], LT_HIoI_RoI[1] - RoI["matrix"][0]] for LT_HIoI_RoI in LT_HIoIs_in_RoI
        ]
        UT_HIoIs_in_RoI_proj = [
            [UT_HIoI_RoI[0] - RoI["matrix"][0], UT_HIoI_RoI[1] - RoI["matrix"][0]] for UT_HIoI_RoI in UT_HIoIs_in_RoI
        ]
        LT_VIoIs_in_RoI_proj = [
            [LT_VIoI_RoI[0] - RoI["matrix"][0], LT_VIoI_RoI[1] - RoI["matrix"][0]] for LT_VIoI_RoI in LT_VIoIs_in_RoI
        ]
        UT_VIoIs_in_RoI_proj = [
            [UT_VIoI_RoI[0] - RoI["matrix"][0], UT_VIoI_RoI[1] - RoI["matrix"][0]] for UT_VIoI_RoI in UT_VIoIs_in_RoI
        ]

        if constrain_height:
            LT_peaks_ids_RoI = [
                LT_peaks_ids_in_candida_in_RoI
                for n, LT_peaks_ids_in_candida_in_RoI in enumerate(LT_peaks_ids)
                if n in ids_LT_MPs_in_RoI
            ]
            LT_peaks_ids_RoI_proj = [
                [
                    LT_peak_idx_in_candida_in_RoI - RoI["matrix"][0]
                    for LT_peak_idx_in_candida_in_RoI in LT_peaks_ids_in_candida_in_RoI
                    if 0 < LT_peak_idx_in_candida_in_RoI - RoI["matrix"][0] < Iproc_RoI.shape[0]
                ]
                for LT_peaks_ids_in_candida_in_RoI in LT_peaks_ids_RoI
            ]
            UT_peaks_ids_RoI = [
                UT_peaks_ids_in_candida_in_RoI
                for n, UT_peaks_ids_in_candida_in_RoI in enumerate(UT_peaks_ids)
                if n in ids_UT_MPs_in_RoI
            ]
            UT_peaks_ids_RoI_proj = [
                [
                    UT_peak_idx_in_candida_in_RoI - RoI["matrix"][0]
                    for UT_peak_idx_in_candida_in_RoI in UT_peaks_ids_in_candida_in_RoI
                    if 0 < UT_peak_idx_in_candida_in_RoI - RoI["matrix"][0] < Iproc_RoI.shape[0]
                ]
                for UT_peaks_ids_in_candida_in_RoI in UT_peaks_ids_RoI
            ]

        print("3.4) Plotting candidate stripes restricted to HIoIs...")

        # Plot of the candidate stripes within the RoI:
        IO.plot_stripes(
            Iproc_RoI,
            LT_HIoIs_in_RoI_proj,
            LT_VIoIs_in_RoI_proj,
            [],
            [],
            RoI["genomic"],
            resolution,
            plot_in_bp=True,
            output_folder=output_folder,
            file_name=f"LT_all_candidates.jpg",
            title=None,
            display=False,
        )

        if constrain_height:
            IO.plot_stripes_and_peaks(
                Iproc_RoI,
                LT_HIoIs_in_RoI_proj,
                LT_VIoIs_in_RoI_proj,
                [],
                [],
                LT_peaks_ids_RoI_proj,
                [],
                RoI["genomic"],
                resolution,
                plot_in_bp=True,
                output_folder=output_folder,
                file_name=f"LT_all_candidates_and_peaks.jpg",
                title=None,
                display=False,
            )

        IO.plot_stripes(
            Iproc_RoI,
            [],
            [],
            UT_HIoIs_in_RoI_proj,
            UT_VIoIs_in_RoI_proj,
            RoI["genomic"],
            resolution,
            plot_in_bp=True,
            output_folder=output_folder,
            file_name=f"UT_all_candidates.jpg",
            title=None,
            display=False,
        )

        if constrain_height:
            IO.plot_stripes_and_peaks(
                Iproc_RoI,
                [],
                [],
                UT_HIoIs_in_RoI_proj,
                UT_VIoIs_in_RoI_proj,
                [],
                UT_peaks_ids_RoI_proj,
                RoI["genomic"],
                resolution,
                plot_in_bp=True,
                output_folder=output_folder,
                file_name=f"UT_all_candidates_and_peaks.jpg",
                title=None,
                display=False,
            )

    print("3.6) Bar plots of widths and heights...")
    LT_widths = [HIoI[1] - HIoI[0] for HIoI in LT_HIoIs]
    LT_heights = [VIoI[1] - VIoI[0] for VIoI in LT_VIoIs]
    UT_widths = [HIoI[1] - HIoI[0] for HIoI in UT_HIoIs]
    UT_heights = [VIoI[1] - VIoI[0] for VIoI in UT_VIoIs]
    if be_verbose and output_folder is not None:
        fig, ax = plt.subplots(1, 1)
        sns.histplot(
            data=pd.DataFrame(LT_widths),
            kde=False,
            legend=False,
            fill=True,
            discrete=True,
            color="#2F539B",
            edgecolor=None,
            alpha=1,
        )
        plt.xlim(0, max(max(LT_widths), max(UT_widths)) + 1)
        plt.title("Widths")
        plt.savefig(f"{output_folder}/LT_histogram_widths.jpg", bbox_inches="tight")
        plt.close()
        fig, ax = plt.subplots(1, 1)
        sns.histplot(
            data=pd.DataFrame(UT_widths),
            kde=False,
            legend=False,
            fill=True,
            discrete=True,
            color="#2F539B",
            edgecolor=None,
            alpha=1,
        )
        plt.xlim(0, max(max(LT_widths), max(UT_widths)) + 1)
        plt.title("Widths")
        plt.savefig(f"{output_folder}/UT_histogram_widths.jpg", bbox_inches="tight")
        plt.close()
        fig, ax = plt.subplots(1, 1)
        sns.histplot(
            data=pd.DataFrame(LT_heights),
            kde=False,
            legend=False,
            fill=True,
            discrete=True,
            color="#2F539B",
            edgecolor=None,
            alpha=1,
        )
        plt.xlim(0, max(max(LT_widths), max(UT_heights)) + 1)
        plt.title("Heights")
        plt.savefig(f"{output_folder}/LT_histogram_heights.jpg", bbox_inches="tight")
        plt.close()
        fig, ax = plt.subplots(1, 1)
        sns.histplot(
            data=pd.DataFrame(UT_heights),
            kde=False,
            legend=False,
            fill=True,
            discrete=True,
            color="#2F539B",
            edgecolor=None,
            alpha=1,
        )
        plt.xlim(0, max(max(LT_heights), max(UT_heights)) + 1)
        plt.title("Heights")
        plt.savefig(f"{output_folder}/UT_histogram_heights.jpg", bbox_inches="tight")
        plt.close()

    return result


def step_4(
    result: IO.Result,
    L,
    U,
    resolution=None,
    thresholds_relative_change=None,
    Iproc_RoI=None,
    RoI=None,
    output_folder=None,
):
    if result.empty:
        print("4) No candidates found by step 2. Returning immediately!")
        return result

    print("4.1) Computing and saving biological descriptors")
    for LT_candidate_stripe in result.get("stripes", "LT"):
        LT_candidate_stripe.compute_biodescriptors(L)
    for UT_candidate_stripe in result.get("stripes", "UT"):
        UT_candidate_stripe.compute_biodescriptors(U)

    if all(param is not None for param in [resolution, thresholds_relative_change, Iproc_RoI, RoI, output_folder]):

        print("4.2) Thresholding...")

        # Retrieve data:
        LT_MPs = [c_s.seed for c_s in result.get("stripes", "LT")]
        UT_MPs = [c_s.seed for c_s in result.get("stripes", "UT")]
        LT_HIoIs = [[c_s.left_bound, c_s.right_bound] for c_s in result.get("stripes", "LT")]
        LT_VIoIs = [[c_s.top_bound, c_s.bottom_bound] for c_s in result.get("stripes", "LT")]
        UT_HIoIs = [[c_s.left_bound, c_s.right_bound] for c_s in result.get("stripes", "UT")]
        UT_VIoIs = [[c_s.top_bound, c_s.bottom_bound] for c_s in result.get("stripes", "UT")]

        for threshold in thresholds_relative_change:

            # Filtration:
            LT_candidates2keep = [
                index
                for index, rel_change in enumerate(s.rel_change for s in result.get("stripes", "LT"))
                if rel_change >= threshold
            ]
            UT_candidates2keep = [
                index
                for index, rel_change in enumerate(s.rel_change for s in result.get("stripes", "UT"))
                if rel_change >= threshold
            ]

            LT_filt_MPs = [LT_MPs[num_cand] for num_cand in LT_candidates2keep]
            LT_filt_HIoIs = [LT_HIoIs[num_cand] for num_cand in LT_candidates2keep]
            LT_filt_VIoIs = [LT_VIoIs[num_cand] for num_cand in LT_candidates2keep]

            UT_filt_MPs = [UT_MPs[num_cand] for num_cand in UT_candidates2keep]
            UT_filt_HIoIs = [UT_HIoIs[num_cand] for num_cand in UT_candidates2keep]
            UT_filt_VIoIs = [UT_VIoIs[num_cand] for num_cand in UT_candidates2keep]

            # Plotting stripes in range:
            if RoI is not None:
                LT_candidates2keep_in_RoI = np.where(
                    (np.array(LT_filt_MPs) > RoI["matrix"][0]) & (np.array(LT_filt_MPs) < RoI["matrix"][1])
                )[0]
                LT_filt_MPs_in_RoI = np.array(LT_filt_MPs)[LT_candidates2keep_in_RoI].tolist()
                LT_filt_MPs_in_RoI_proj = [a - RoI["matrix"][0] for a in LT_filt_MPs_in_RoI]
                LT_filt_HIoIs_in_RoI = np.array(LT_filt_HIoIs)[LT_candidates2keep_in_RoI].tolist()
                LT_filt_HIoIs_in_RoI_proj = [
                    [a[0] - RoI["matrix"][0], a[1] - RoI["matrix"][2]] for a in LT_filt_HIoIs_in_RoI
                ]
                LT_filt_VIoIs_in_RoI = np.array(LT_filt_VIoIs)[LT_candidates2keep_in_RoI].tolist()
                LT_filt_VIoIs_in_RoI_proj = [
                    [a[0] - RoI["matrix"][0], a[1] - RoI["matrix"][2]] for a in LT_filt_VIoIs_in_RoI
                ]

                IO.plot_stripes(
                    Iproc_RoI,
                    LT_filt_HIoIs_in_RoI_proj,
                    LT_filt_VIoIs_in_RoI_proj,
                    [],
                    [],
                    RoI["genomic"],
                    resolution,
                    plot_in_bp=True,
                    output_folder=f"{output_folder}",
                    file_name=f"LT_{threshold:.2f}.jpg",
                    title=None,
                    display=False,
                )

                UT_candidates2keep_in_RoI = np.where(
                    (np.array(UT_filt_MPs) > RoI["matrix"][0]) & (np.array(UT_filt_MPs) < RoI["matrix"][1])
                )[0]
                UT_filt_MPs_in_RoI = np.array(UT_filt_MPs)[UT_candidates2keep_in_RoI].tolist()
                UT_filt_MPs_in_RoI_proj = [a - RoI["matrix"][0] for a in UT_filt_MPs_in_RoI]
                UT_filt_HIoIs_in_RoI = np.array(UT_filt_HIoIs)[UT_candidates2keep_in_RoI].tolist()
                UT_filt_HIoIs_in_RoI_proj = [
                    [a[0] - RoI["matrix"][0], a[1] - RoI["matrix"][2]] for a in UT_filt_HIoIs_in_RoI
                ]
                UT_filt_VIoIs_in_RoI = np.array(UT_filt_VIoIs)[UT_candidates2keep_in_RoI].tolist()
                UT_filt_VIoIs_in_RoI_proj = [
                    [a[0] - RoI["matrix"][0], a[1] - RoI["matrix"][2]] for a in UT_filt_VIoIs_in_RoI
                ]

                IO.plot_stripes(
                    Iproc_RoI,
                    [],
                    [],
                    UT_filt_HIoIs_in_RoI_proj,
                    UT_filt_VIoIs_in_RoI_proj,
                    RoI["genomic"],
                    resolution,
                    plot_in_bp=True,
                    output_folder=f"{output_folder}",
                    file_name=f"UT_{threshold:.2f}.jpg",
                    title=None,
                    display=False,
                )

    return result



# Copyright (C) 2024 Andrea Raffo <andrea.raffo@ibv.uio.no>
#
# SPDX-License-Identifier: MIT

import os

import h5py
import hictkpy
import numpy as np

from . import IO


def _raise_invalid_bin_type_except(f: hictkpy.File):
    raise RuntimeError(f"Only files with a uniform bin size are supported, found \"{f.attributes()['bin-type']}\".")


def cmap_loading(path: os.PathLike, resolution: int):
    try:
        if not isinstance(resolution, int):
            raise TypeError("resolution must be an integer.")

        if resolution <= 0:
            raise ValueError("resolution must be greater than zero.")

        if hictkpy.is_scool_file(path):
            raise RuntimeError(".scool files are not currently supported.")
        if hictkpy.is_cooler(path):
            f = hictkpy.File(path)
            if f.resolution() == 0:
                _raise_invalid_bin_type_except(f)
            if f.resolution() != resolution:
                raise RuntimeError(f"expected {resolution} resolution, found {f.resolution()}.")
        else:
            f = hictkpy.MultiResFile(path)[resolution]
    except RuntimeError as e:
        raise RuntimeError(f'error opening file "{path}"') from e

    if f.attributes().get("bin-type", "fixed") != "fixed":
        _raise_invalid_bin_type_except(f)

    # Retrieve metadata:
    chr_starts = [0]  # left ends of each chromosome  (in matrix coordinates)
    chr_ends = []  # right ends of each chromosome (in matrix coordinates)
    chr_sizes = []  # integer bp lengths, one per chromosome
    for bp_length in f.chromosomes().values():
        chr_sizes.append(bp_length)
        chr_ends.append(chr_starts[-1] + int(np.ceil(bp_length / resolution)))
        chr_starts.append(chr_ends[-1])
        # print(f"{chr_starts[-2]}-{chr_ends[-1]}-{chr_sizes[-1]}")
    chr_starts.pop(-1)

    return f, chr_starts, chr_ends, chr_sizes


def chromosomes_to_study(chromosomes, length_in_bp, min_size_allowed):
    # Extract the list of chromosomes:
    chr_ids = list(range(len(chromosomes)))
    c_pairs = list(zip(chr_ids, chromosomes))

    # Remove overly-short chromosomes:
    surviving_indices = [i for i, e in enumerate(length_in_bp) if e > min_size_allowed]
    deleted_indices = [i for i, e in enumerate(length_in_bp) if e <= min_size_allowed]
    if len(deleted_indices) > 0:
        print(
            f"{IO.ANSI.RED}ATT: The following chromosomes are discarded because shorter than --min-chrom-size = "
            f"{min_size_allowed} bp: {[chromosomes[i] for i in deleted_indices]}{IO.ANSI.ENDC}"
        )
        c_pairs = [c_pairs[i] for i in surviving_indices]

        # If there is no chromosome left, exit:
        if len(c_pairs) == 0:
            raise ValueError(f"\nNo chromosome is long enough... decrease the parameter --min-chrom-size")

    return c_pairs


def define_RoI(where_roi, chr_start, chr_end, resolution):
    # Region of Interest (RoI) in genomic and matrix coordinates:
    if where_roi == "middle":
        RoI_length = 2000000
        e1 = ((chr_end - chr_start) * resolution - RoI_length) / 2
        e2 = e1 + RoI_length
        RoI = dict()
        RoI["genomic"] = [int(e1), int(e2), int(e1), int(e2)]  # genomic coordinates
        RoI["matrix"] = [int(roi / resolution) for roi in RoI["genomic"]]  # matrix coordinates
    elif where_roi == "start":
        RoI_length = 2000000
        e1 = 0
        e2 = e1 + RoI_length
        RoI = dict()
        RoI["genomic"] = [int(e1), int(e2), int(e1), int(e2)]  # genomic coordinates
        RoI["matrix"] = [int(roi / resolution) for roi in RoI["genomic"]]  # matrix coordinates
    else:
        RoI = None

    return RoI


# Define a function to visit groups and save terminal group names in a list
def save_terminal_groups(name, obj):
    group_names = []
    if isinstance(obj, h5py.Group):
        has_subgroups = any(isinstance(child_obj, h5py.Group) for _, child_obj in obj.items())
        if not has_subgroups:
            group_names.append(name)
    return group_names


# Define a function to visit datasets and save their names in a list
def save_all_datasets(name, obj):
    dataset_names = []
    if isinstance(obj, h5py.Dataset):  # check if obj is a group or dataset
        dataset_names.append(name)
    return dataset_names



# Copyright (C) 2024 Andrea Raffo <andrea.raffo@ibv.uio.no>
#
# SPDX-License-Identifier: MIT

import contextlib
import multiprocessing as mp
import time
from typing import Any, Dict

import numpy as np

from stripepy import IO, others, stripepy


def _generate_metadata_attribute(configs_input: Dict[str, Any], configs_thresholds: Dict[str, Any]) -> Dict[str, Any]:
    return {
        "constrain-heights": configs_thresholds["constrain_heights"],
        "genomic-belt": configs_input["genomic_belt"],
        "global-persistence-minimum": configs_thresholds["glob_pers_min"],
        "local-persistence-minimum": configs_thresholds["loc_pers_min"],
        "local-trend-minimum": configs_thresholds["loc_trend_min"],
        "max-width": configs_thresholds["max_width"],
        "min-chromosome-size": configs_thresholds["min_chrom_size"],
    }


def run(
    configs_input: Dict[str, Any],
    configs_thresholds: Dict[str, Any],
    configs_output: Dict[str, Any],
    configs_other: Dict[str, Any],
):
    # How long does stripepy take to analyze the whole Hi-C matrix?
    start_global_time = time.time()

    # Data loading:
    f, chr_starts, chr_ends, bp_lengths = others.cmap_loading(configs_input["contact_map"], configs_input["resolution"])

    # Remove existing folders:
    # configs_output["output_folder"] = (
    #     f"{configs_output['output_folder']}/{configs_input['contact_map'].stem}/{configs_input['resolution']}"
    # )
    configs_output["output_folder"] = (
        configs_output["output_folder"] / configs_input["contact_map"].stem / str(configs_input["resolution"])
    )
    IO.remove_and_create_folder(configs_output["output_folder"], configs_output["force"])

    # Extract a list of tuples where each tuple is (index, chr), e.g. (2,'chr3'):
    c_pairs = others.chromosomes_to_study(
        list(f.chromosomes().keys()), bp_lengths, configs_thresholds["min_chrom_size"]
    )

    with contextlib.ExitStack() as ctx:
        # Create HDF5 file to store candidate stripes:
        h5 = ctx.enter_context(IO.ResultFile(configs_output["output_folder"] / "results.hdf5", "w"))

        h5.init_file(f, configs_input["normalization"], _generate_metadata_attribute(configs_input, configs_thresholds))

        # Set up the process pool when appropriate
        if configs_other["nproc"] > 1:
            pool = ctx.enter_context(mp.Pool(configs_other["nproc"]))
        else:
            pool = None

        # Lopping over all chromosomes:
        for this_chr_idx, this_chr in c_pairs:

            print(f"\n{IO.ANSI.RED}CHROMOSOME {this_chr}{IO.ANSI.ENDC}")
            start_local_time = time.time()

            # Removing and creating folders to store output files:
            # configs_input['roi'] = None
            if configs_input["roi"] is not None:
                IO.create_folders_for_plots(configs_output["output_folder"] / "plots" / this_chr)

            I = f.fetch(this_chr, normalization=configs_input["normalization"]).to_csr("full")

            # RoI:
            RoI = others.define_RoI(
                configs_input["roi"], chr_starts[this_chr_idx], chr_ends[this_chr_idx], configs_input["resolution"]
            )
            print(f"RoI is: {RoI}")

            print(f"{IO.ANSI.YELLOW}Step 1: pre-processing step{IO.ANSI.ENDC}")
            start_time = time.time()
            if all(param is not None for param in [RoI, configs_output["output_folder"]]):
                output_folder_1 = f"{configs_output['output_folder']}/plots/{this_chr}/1_preprocessing/"
                LT_Iproc, UT_Iproc, Iproc_RoI = stripepy.step_1(
                    I,
                    configs_input["genomic_belt"],
                    configs_input["resolution"],
                    RoI=RoI,
                    output_folder=output_folder_1,
                )
            else:
                LT_Iproc, UT_Iproc, _ = stripepy.step_1(I, configs_input["genomic_belt"], configs_input["resolution"])
                Iproc_RoI = None
            print(f"Execution time of step 1: {time.time() - start_time} seconds ---")

            # Find the indices where the sum is zero
            # TODO: DO SOMETHING
            # zero_indices = np.where(np.sum(Iproc_RoI, axis=0) == 0)[0]
            # print(np.min(np.sum(LT_Iproc + UT_Iproc, axis=0)))
            # print(np.max(np.sum(LT_Iproc + UT_Iproc, axis=0)))
            # np.savetxt("trend.txt", np.sum(LT_Iproc + UT_Iproc, axis=0))
            # exit()

            print(f"{IO.ANSI.YELLOW}Step 2: Topological Data Analysis{IO.ANSI.ENDC}")
            start_time = time.time()
            if all(param is not None for param in [Iproc_RoI, RoI, configs_output["output_folder"]]):
                output_folder_2 = f"{configs_output['output_folder']}/plots/{this_chr}/2_TDA/"
                result = stripepy.step_2(
                    this_chr,
                    LT_Iproc,
                    UT_Iproc,
                    configs_input["resolution"],
                    configs_thresholds["glob_pers_min"],
                    Iproc_RoI=Iproc_RoI,
                    RoI=RoI,
                    output_folder=output_folder_2,
                )
            else:
                result = stripepy.step_2(
                    this_chr,
                    LT_Iproc,
                    UT_Iproc,
                    configs_input["resolution"],
                    configs_thresholds["glob_pers_min"],
                )
            print(f"Execution time of step 2: {time.time() - start_time} seconds ---")

            print(f"{IO.ANSI.YELLOW}Step 3: Shape analysis{IO.ANSI.ENDC}")
            start_time = time.time()

            if all(param is not None for param in [Iproc_RoI, RoI, configs_output["output_folder"]]):
                output_folder_3 = f"{configs_output['output_folder']}/plots/{this_chr}/3_shape_analysis/"
                result = stripepy.step_3(
                    result,
                    LT_Iproc,
                    UT_Iproc,
                    configs_input["resolution"],
                    configs_input["genomic_belt"],
                    configs_thresholds["max_width"],
                    configs_thresholds["constrain_heights"],
                    configs_thresholds["loc_pers_min"],
                    configs_thresholds["loc_trend_min"],
                    Iproc_RoI=Iproc_RoI,
                    RoI=RoI,
                    output_folder=output_folder_3,
                    map=pool.map if pool is not None else map,
                )
            else:
                result = stripepy.step_3(
                    result,
                    LT_Iproc,
                    UT_Iproc,
                    configs_input["resolution"],
                    configs_input["genomic_belt"],
                    configs_thresholds["max_width"],
                    configs_thresholds["constrain_heights"],
                    configs_thresholds["loc_pers_min"],
                    configs_thresholds["loc_trend_min"],
                    map=pool.map if pool is not None else map,
                )

            print(f"Execution time of step 3: {time.time() - start_time} seconds ---")

            print(f"{IO.ANSI.YELLOW}Step 4: Statistical analysis and post-processing{IO.ANSI.ENDC}")
            start_time = time.time()

            if all(param is not None for param in [Iproc_RoI, RoI, configs_output["output_folder"]]):
                output_folder_4 = f"{configs_output['output_folder']}/plots/{this_chr}/4_biological_analysis/"
                thresholds_relative_change = np.arange(0.0, 15.2, 0.2)
                result = stripepy.step_4(
                    result,
                    LT_Iproc,
                    UT_Iproc,
                    configs_input["resolution"],
                    thresholds_relative_change,
                    Iproc_RoI=Iproc_RoI,
                    RoI=RoI,
                    output_folder=output_folder_4,
                )
            else:
                result = stripepy.step_4(result, LT_Iproc, UT_Iproc)

            print(f"Execution time of step 4: {time.time() - start_time} seconds ---")

            print(f'Writing results for "{this_chr}" to file "{h5.path}"...')
            h5.write_descriptors(result)

            print(f"{IO.ANSI.CYAN}This chromosome has taken {(time.time() - start_local_time)} seconds{IO.ANSI.ENDC}")

    print(f"\n\n{IO.ANSI.RED}The code has run for {(time.time() - start_global_time) / 60} minutes{IO.ANSI.ENDC}")



# Copyright (C) 2024 Andrea Raffo <andrea.raffo@ibv.uio.no>
#
# SPDX-License-Identifier: MIT

import argparse
import math
import multiprocessing as mp
import pathlib
from importlib.metadata import version
from typing import Any, Dict, List, Tuple


# Create a custom formatter to allow multiline and bulleted descriptions
class CustomFormatter(argparse.RawTextHelpFormatter):
    def _fill_text(self, text, width, indent):
        return "".join([indent + line + "\n" for line in text.splitlines()])


def _num_cpus(arg: str) -> int:
    try:
        n = int(arg)
        if 0 < n <= mp.cpu_count():
            return n
    except:  # noqa
        pass

    raise ValueError(f"Not a valid number of CPU cores (allowed values are integers between 1 and {mp.cpu_count()})")


def _existing_file(arg: str) -> pathlib.Path:
    if (path := pathlib.Path(arg)).is_file():
        return path

    raise FileNotFoundError(arg)


def _output_dir_checked(arg: str) -> pathlib.Path:
    parent = pathlib.Path(arg).parent
    if parent.exists() and parent.is_dir():
        return pathlib.Path(arg)

    raise FileNotFoundError(f'Output folder "{arg}" is not reachable: parent folder does not exist')


def _probability(arg) -> float:
    if 0 <= (n := float(arg)) <= 1:
        return n

    raise ValueError("Not a valid probability")


def _positive_float(arg) -> float:
    if (n := float(arg)) > 0:
        return n

    raise ValueError("Not a positive float")


def _positive_int(arg) -> float:
    if (n := int(arg)) > 0:
        return n

    raise ValueError("Not a positive int")


def _make_stripepy_call_subcommand(main_parser) -> argparse.ArgumentParser:
    sc: argparse.ArgumentParser = main_parser.add_parser(
        "call",
        help="stripepy works in four consecutive steps: \n"
        "• Step 1: Pre-processing\n"
        "• Step 2: Recognition of loci of interest (also called 'seeds')\n"
        "• Step 3: Shape analysis (i.e., width and height estimation)\n"
        "• Step 4: Signal analysis\n",
    )

    sc.add_argument(
        "contact-map",
        type=_existing_file,
        help="Path to a .cool, .mcool, or .hic file for input.",
    )

    sc.add_argument(
        "resolution",
        type=int,
        help="Resolution (in bp).",
    )

    sc.add_argument(
        "-n",
        "--normalization",
        type=str,
        help="Normalization to fetch (default: 'None').",
    )

    sc.add_argument(
        "-b",
        "--genomic-belt",
        type=int,
        default=5_000_000,
        help="Radius of the band, centred around the diagonal, where the search is restricted to (in bp, default: 5000000).",
    )

    sc.add_argument(
        "--roi",
        type=str,
        default=None,
        help="Specify 'middle' or input range as 'chr2:10000000-12000000' (default: None)",
    )

    sc.add_argument(
        "-o",
        "--output-folder",
        type=_output_dir_checked,
        default=pathlib.Path("."),
        help="Path to the folder where the user wants the output to be placed (default: current folder).",
    )

    sc.add_argument(
        "--max-width",
        type=int,
        default=100_000,
        help="Maximum stripe width, in bp.",
    )

    sc.add_argument(
        "--glob-pers-min",
        type=_probability,
        default=0.05,
        help="Threshold value between 0 and 1 to filter persistence maxima points and identify loci of interest, "
        "aka seeds (default: 0.2).",
    )

    sc.add_argument(
        "--constrain-heights",
        action="store_true",
        default=False,
        help="Use peaks in signal to constrain the stripe height (default: 'False')",
    )

    sc.add_argument(
        "--loc-pers-min",
        type=_probability,
        default=0.33,
        help="Threshold value between 0 and 1 to find peaks in signal in a horizontal domain while estimating the "
        "height of a stripe; when --constrain-heights is set to 'False', it is not used (default: 0.2).",
    )

    sc.add_argument(
        "--loc-trend-min",
        type=_probability,
        default=0.25,
        help="Threshold value between 0 and 1 to estimate the height of a stripe (default: 0.1); "
        "the higher this value, the shorter the stripe; it is always used when --constrain-heights is set to "
        "'False', but could be necessary also when --constrain-heights is 'True' and no persistent maximum other "
        "than the global maximum is found.",
    )

    sc.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help="Overwrite existing file(s).",
    )

    sc.add_argument(
        "-p",
        "--nproc",
        type=_num_cpus,
        default=1,
        help="Maximum number of parallel processes to use.",
    )

    sc.add_argument(
        "--min-chrom-size",
        type=int,
        default=2_000_000,
        help="Minimum size, in bp, for a chromosome to be analysed (default: 2 Mbp).",
    )

    return sc


def _make_stripepy_download_subcommand(main_parser) -> argparse.ArgumentParser:
    sc: argparse.ArgumentParser = main_parser.add_parser(
        "download",
        help="Helper command to simplify downloading datasets that can be used to test StripePy.",
    )

    def get_avail_ref_genomes():
        from .download import _get_datasets

        return {record["assembly"] for record in _get_datasets(math.inf).values() if "assembly" in record}

    grp = sc.add_mutually_exclusive_group(required=False)
    grp.add_argument(
        "--assembly",
        type=str,
        choices=get_avail_ref_genomes(),
        help="Restrict downloads to the given reference genome assembly.",
    )
    grp.add_argument(
        "--name",
        type=str,
        help="Name of the dataset to be downloaded.\n"
        "When not provided, randomly select and download a dataset based on the provided CLI options (if any).",
    )
    grp.add_argument(
        "--list-only",
        action="store_true",
        default=False,
        help="Print the list of available datasets and return.",
    )

    sc.add_argument(
        "--max-size",
        type=_positive_float,
        default=512.0,
        help="Upper bound for the size of the files to be considered when --name is not provided.",
    )
    sc.add_argument(
        "-o",
        "--output",
        type=pathlib.Path,
        dest="output_path",
        help="Path where to store the downloaded file.",
    )
    sc.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help="Overwrite existing file(s).",
    )

    return sc


def _make_stripepy_plot_subcommand(main_parser) -> argparse.ArgumentParser:
    subparser = main_parser.add_parser(
        "plot",
        help="Generate various static plots useful to visually inspect the output produced by stripepy call.",
    ).add_subparsers(title="plot_subcommands", dest="plot_type", required=True, help="List of available subcommands:")

    def add_common_options(sc, region_is_randomized: bool):
        sc.add_argument(
            "output-name",
            type=pathlib.Path,
            help="Path where to store the generated plot.",
        )

        region_help_msg1 = (
            "Genomic region to be plotted (UCSC format). When not specified, a random 2.5Mb region is plotted."
        )
        region_help_msg2 = (
            "Genomic region to be plotted (UCSC format). When not specified, data for the entire genome is plotted."
        )

        sc.add_argument(
            "--region",
            type=str,
            help=region_help_msg1 if region_is_randomized else region_help_msg2,
        )
        sc.add_argument(
            "--dpi",
            type=_positive_int,
            default=300,
            help="DPI of the generated plot (ignored when the output format is a vector graphic).",
        )
        sc.add_argument(
            "--seed",
            type=int,
            default=7606490399616306585,
            help="Seed for random number generation.",
        )
        sc.add_argument(
            "-n",
            "--normalization",
            type=str,
            help="Normalization to fetch (default: 'None').",
        )
        sc.add_argument(
            "-f",
            "--force",
            action="store_true",
            default=False,
            help="Overwrite existing file(s).",
        )

    def add_stripepy_hdf5_option(sc):
        sc.add_argument(
            "stripepy-hdf5",
            type=_existing_file,
            help="Path to the .hdf5 generated by stripepy call.",
        )

    sc = subparser.add_parser(
        "contact-map",
        help="Plot stripes and other features over the Hi-C matrix.",
        aliases=["cm"],
    )
    sc.add_argument(
        "contact-map",
        type=_existing_file,
        help="Path to the .cool, .mcool, or .hic file used to call stripes.",
    )
    sc.add_argument(
        "resolution",
        type=_positive_int,
        help="Resolution (in bp).",
    )

    sc.add_argument(
        "--stripepy-hdf5",
        type=_existing_file,
        help="Path to the .hdf5 generated by stripepy call. Required when highlighting stripes or seeds.",
    )

    sc.add_argument(
        "--relative-change-threshold",
        type=_positive_float,
        help="Cutoff for the relative change.\n"
        "Only used when highlighting architectural stripes.\n"
        "The relative change is computed as the ratio between the average number of interactions "
        "found inside a stripe and the number of interactions in a neighborhood outside of the stripe.",
    )

    grp = sc.add_mutually_exclusive_group()
    grp.add_argument(
        "--highlight-seeds",
        action="store_true",
        default=False,
        help="Highlight the stripe seeds.",
    )
    grp.add_argument(
        "--highlight-stripes",
        action="store_true",
        default=False,
        help="Highlight the architectural stripes.",
    )
    sc.add_argument(
        "--ignore-stripe-heights",
        action="store_true",
        default=False,
        help="Ignore the stripes height. Has no effect when --highlight-stripes is not specified.",
    )
    sc.add_argument(
        "--cmap",
        type=str,
        default="fruit_punch",
        help="Color map used to plot Hi-C interactions.\n"
        "Can be any of the color maps supported by matplotlib as well as: fall, fruit_punch, "
        "blues, acidblues, and nmeth.",
    )

    grp = sc.add_mutually_exclusive_group()
    grp.add_argument(
        "--linear-scale",
        action="store_false",
        dest="log_scale",
        help="Plot interactions in linear scale.",
    )
    grp.add_argument(
        "--log-scale",
        action="store_true",
        dest="log_scale",
        default=True,
        help="Plot interactions in log scale.",
    )

    add_common_options(sc, region_is_randomized=True)

    sc = subparser.add_parser(
        "pseudodistribution",
        help="Plot the pseudo-distribution over the given region of interest.",
        aliases=["pd"],
    )
    add_stripepy_hdf5_option(sc)
    add_common_options(sc, region_is_randomized=True)

    sc = subparser.add_parser(
        "stripe-hist",
        help="Generate and plot the histograms showing the distribution of the stripe heights and widths.",
        aliases=["hist"],
    )
    add_stripepy_hdf5_option(sc)
    add_common_options(sc, region_is_randomized=False)

    return sc


def _make_stripepy_view_subcommand(main_parser) -> argparse.ArgumentParser:
    sc: argparse.ArgumentParser = main_parser.add_parser(
        "view",
        help="Fetch stripes from the HDF5 file produced by stripepy call.",
    )

    sc.add_argument(
        dest="h5_file",
        metavar="h5-file",
        type=_existing_file,
        help="Path to the HDF5 file generated by stripepy call.",
    )

    sc.add_argument(
        "--relative-change-threshold",
        type=float,
        default=5.0,
        help="Cutoff for the relative change.\n"
        "The relative change is computed as the ratio between the average number of interactions\n"
        "found inside a stripe and the number of interactions in a neighborhood outside of the stripe.",
    )

    sc.add_argument(
        "--transform",
        type=str,
        choices=["transpose_to_ut", "transpose_to_lt", None],
        default=None,
        help="Control if and how stripe coordinates should be transformed.",
    )

    return sc


def _make_cli() -> argparse.ArgumentParser:
    cli = argparse.ArgumentParser(
        description="stripepy is designed to recognize linear patterns in contact maps (.hic, .mcool, .cool) "
        "through the geometric reasoning, including topological persistence and quasi-interpolation. ",
        formatter_class=CustomFormatter,
    )

    sub_parser = cli.add_subparsers(
        title="subcommands", dest="subcommand", required=True, help="List of available subcommands:"
    )

    _make_stripepy_call_subcommand(sub_parser)
    _make_stripepy_download_subcommand(sub_parser)
    _make_stripepy_plot_subcommand(sub_parser)
    _make_stripepy_view_subcommand(sub_parser)

    cli.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=version("stripepy-hic")),
    )

    return cli


def _process_stripepy_call_args(args: Dict[str, Any]) -> Dict[str, Any]:

    # Gather input parameters in dictionaries:
    configs_input = {key: args[key] for key in ["contact_map", "resolution", "normalization", "genomic_belt", "roi"]}
    configs_thresholds = {
        key: args[key]
        for key in [
            "glob_pers_min",
            "constrain_heights",
            "loc_pers_min",
            "loc_trend_min",
            "max_width",
            "min_chrom_size",
        ]
    }
    configs_output = {key: args[key] for key in ["output_folder", "force"]}

    configs_other = {"nproc": args["nproc"]}

    # Print the used parameters (chosen or default-ones):
    print("\nArguments:")
    print(f"--contact-map: {configs_input['contact_map']}")
    print(f"--resolution: {configs_input['resolution']}")
    print(f"--normalization: {configs_input['normalization']}")
    print(f"--genomic-belt: {configs_input['genomic_belt']}")
    print(f"--roi: {configs_input['roi']}")
    print(f"--max-width: {configs_thresholds['max_width']}")
    print(f"--glob-pers-min: {configs_thresholds['glob_pers_min']}")
    print(f"--constrain-heights: {configs_thresholds['constrain_heights']}")
    print(f"--loc-pers-min: {configs_thresholds['loc_pers_min']}")
    print(f"--loc-trend-min: {configs_thresholds['loc_trend_min']}")
    print(f"--min-chrom-size: {configs_thresholds['min_chrom_size']}")
    print(f"--output-folder: {configs_output['output_folder']}")
    print(f"--nproc: {configs_other['nproc']}")
    print(f"--force: {configs_output['force']}")

    return {
        "configs_input": configs_input,
        "configs_thresholds": configs_thresholds,
        "configs_output": configs_output,
        "configs_other": configs_other,
    }


def _normalize_args(args: Dict[str, Any]) -> Dict[str, Any]:
    return {k.replace("-", "_"): v for k, v in args.items()}


def parse_args(cli_args: List[str]) -> Tuple[str, Any]:
    # Parse the input parameters:
    args = _normalize_args(vars(_make_cli().parse_args(cli_args)))

    subcommand = args.pop("subcommand")
    if subcommand == "call":
        return subcommand, _process_stripepy_call_args(args)
    if subcommand == "download":
        return subcommand, args
    if subcommand == "plot":
        return subcommand, args
    if subcommand == "view":
        return subcommand, args

    raise NotImplementedError



# Copyright (C) 2024 Roberto Rossini <roberroso@uio.no>
#
# SPDX-License-Identifier: MIT

import contextlib
import hashlib
import io
import json
import pathlib

import pytest

from stripepy import main


def _hash_file(path: pathlib.Path) -> str:
    hasher = hashlib.md5()
    hasher.update(path.read_bytes())
    return hasher.hexdigest()


@pytest.mark.end2end
class TestStripePyDownload:
    @staticmethod
    def test_list_only():
        args = ["download", "--list-only"]
        buff = io.StringIO()
        with contextlib.redirect_stdout(buff):
            main(args)

        data = json.loads(buff.getvalue())
        assert len(data) != 0

    @staticmethod
    def test_download_by_name(tmpdir):
        dest = pathlib.Path(tmpdir) / "out"

        args = ["download", "--name", "__results_v1", "--output", str(dest)]
        main(args)

        assert dest.is_file()

        assert _hash_file(dest) == "632b2a7a6e5c1a24dc3635710ed68a80"

    @staticmethod
    def test_download_random(tmpdir):
        dest = pathlib.Path(tmpdir) / "out"

        args = ["download", "--max-size", "2", "--output", str(dest)]
        main(args)

        assert dest.is_file()

        assert _hash_file(dest) == "d4ab74937dd9062efe4b2acc6ebc8780"



# Copyright (C) 2024 Andrea Raffo <andrea.raffo@ibv.uio.no>
#
# SPDX-License-Identifier: MIT

import numpy as np


def _compute_wQISA_predictions(Y, k):

    # Control points of a weighted quasi-interpolant spline approximation with a k-NN weight function:
    kernel = np.full((k,), 1 / k)
    radius = (k - 1) // 2
    Y_ext = np.pad(Y.tolist(), (radius, radius), mode="edge")
    predictions = np.convolve(Y_ext, kernel, "valid")

    return predictions



# Copyright (C) 2024 Roberto Rossini <roberros@uio.no>
#
# SPDX-License-Identifier: MIT

import pathlib
import shutil
import tarfile

import pytest

from stripepy import main

testdir = pathlib.Path(__file__).resolve().parent.parent


def compare_images(expected: pathlib.Path, actual: pathlib.Path, tol: float = 0.1):
    import matplotlib.testing.compare

    res = matplotlib.testing.compare.compare_images(str(expected), str(actual), tol)
    if res is not None:
        pytest.fail(res)


def extract_image(
    name: str, dest_dir: pathlib.Path, prefix: pathlib.Path = pathlib.Path("stripepy-plot-test-images")
) -> pathlib.Path:
    dest = dest_dir / name
    with tarfile.TarFile.open(testdir / "data" / "stripepy-plot-test-images.tar.xz") as tar:
        with tar.extractfile((prefix / name).as_posix()) as fin, dest.open("wb") as fout:
            shutil.copyfileobj(fin, fout)  # noqa

    return dest_dir / name


@pytest.mark.end2end
class TestStripePyPlot:
    @staticmethod
    def setup_class():
        test_files = [
            testdir / "data" / "4DNFI9GMP2J8.mcool",
            testdir / "data" / "results_4DNFI9GMP2J8_v1.hdf5",
            testdir / "data" / "stripepy-plot-test-images.tar.xz",
        ]

        for f in test_files:
            if not f.exists():
                raise RuntimeError(
                    f'unable to find file "{f}". Did you download the test files prior to running pytest?'
                )

    @staticmethod
    def test_contact_map(tmpdir):
        tmpdir = pathlib.Path(tmpdir)

        matrix_file = testdir / "data" / "4DNFI9GMP2J8.mcool"
        expected = extract_image("contact_map.png", tmpdir)

        resolution = 10_000
        region = "chr2:120100000-122100000"

        outfile = tmpdir / "img.png"
        args = ["plot", "contact-map", str(matrix_file), str(resolution), str(outfile), "--region", region]
        main(args)

        assert outfile.is_file()
        compare_images(expected, outfile)

    @staticmethod
    def test_contact_map_with_seeds(tmpdir):
        tmpdir = pathlib.Path(tmpdir)

        matrix_file = testdir / "data" / "4DNFI9GMP2J8.mcool"
        stripe_file = testdir / "data" / "results_4DNFI9GMP2J8_v1.hdf5"
        expected = extract_image("contact_map_with_seeds.png", tmpdir)

        resolution = 10_000
        region = "chr2:120100000-122100000"

        outfile = tmpdir / "img.png"
        args = [
            "plot",
            "contact-map",
            str(matrix_file),
            str(resolution),
            str(outfile),
            "--stripepy-hdf5",
            str(stripe_file),
            "--highlight-seeds",
            "--region",
            region,
        ]
        main(args)

        assert outfile.is_file()
        compare_images(expected, outfile)

    @staticmethod
    def test_contact_map_with_stripes(tmpdir):
        tmpdir = pathlib.Path(tmpdir)

        matrix_file = testdir / "data" / "4DNFI9GMP2J8.mcool"
        stripe_file = testdir / "data" / "results_4DNFI9GMP2J8_v1.hdf5"
        expected = extract_image("contact_map_with_stripes.png", tmpdir)

        resolution = 10_000
        region = "chr2:120100000-122100000"

        outfile = tmpdir / "img.png"
        args = [
            "plot",
            "contact-map",
            str(matrix_file),
            str(resolution),
            str(outfile),
            "--stripepy-hdf5",
            str(stripe_file),
            "--highlight-stripes",
            "--region",
            region,
        ]
        main(args)

        assert outfile.is_file()
        compare_images(expected, outfile)

    @staticmethod
    def test_contact_map_with_stripes_no_heights(tmpdir):
        tmpdir = pathlib.Path(tmpdir)

        matrix_file = testdir / "data" / "4DNFI9GMP2J8.mcool"
        stripe_file = testdir / "data" / "results_4DNFI9GMP2J8_v1.hdf5"
        expected = extract_image("contact_map_with_stripes_no_heights.png", tmpdir)

        resolution = 10_000
        region = "chr2:120100000-122100000"

        outfile = tmpdir / "img.png"
        args = [
            "plot",
            "contact-map",
            str(matrix_file),
            str(resolution),
            str(outfile),
            "--stripepy-hdf5",
            str(stripe_file),
            "--highlight-stripes",
            "--ignore-stripe-heights",
            "--region",
            region,
        ]
        main(args)

        assert outfile.is_file()
        compare_images(expected, outfile)

    @staticmethod
    def test_pseudodistribution(tmpdir):
        tmpdir = pathlib.Path(tmpdir)

        stripe_file = testdir / "data" / "results_4DNFI9GMP2J8_v1.hdf5"
        expected = extract_image("pseudodistribution.png", tmpdir)

        region = "chr2:120100000-122100000"

        outfile = tmpdir / "img.png"
        args = ["plot", "pseudodistribution", str(stripe_file), str(outfile), "--region", region]
        main(args)

        assert outfile.is_file()
        compare_images(expected, outfile)

    @staticmethod
    def test_stripe_hist(tmpdir):
        tmpdir = pathlib.Path(tmpdir)

        stripe_file = testdir / "data" / "results_4DNFI9GMP2J8_v1.hdf5"
        expected = extract_image("stripe_hist.png", tmpdir)

        region = "chr2:120100000-122100000"

        outfile = tmpdir / "img.png"
        args = ["plot", "stripe-hist", str(stripe_file), str(outfile), "--region", region]
        main(args)

        assert outfile.is_file()
        compare_images(expected, outfile)

    @staticmethod
    def test_stripe_hist_gw(tmpdir):
        tmpdir = pathlib.Path(tmpdir)

        stripe_file = testdir / "data" / "results_4DNFI9GMP2J8_v1.hdf5"
        expected = extract_image("stripe_hist_gw.png", tmpdir)

        outfile = tmpdir / "img.png"
        args = ["plot", "stripe-hist", str(stripe_file), str(outfile)]
        main(args)

        assert outfile.is_file()
        compare_images(expected, outfile)



# Copyright (C) 2024 Roberto Rossini <roberros@uio.no>
#
# SPDX-License-Identifier: MIT

import pathlib
import re
import sys
import warnings
from typing import Union

import numpy as np
import pandas as pd

from stripepy.IO import ResultFile


def _normalize_df(df: pd.DataFrame) -> pd.DataFrame:
    new_cols = []
    for col in df.columns:
        new_cols.append(re.sub(r"\W+", "_", col.lower()))

    df.columns = new_cols
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return df.convert_dtypes()


def _read_stripes(f: ResultFile, chrom: str) -> pd.DataFrame:
    try:
        bio_lt = f.get(chrom, "bio_descriptors", "LT")[["rel_change"]]
        bio_ut = f.get(chrom, "bio_descriptors", "UT")[["rel_change"]]
        geo_lt = f.get(chrom, "geo_descriptors", "LT")
        geo_ut = f.get(chrom, "geo_descriptors", "UT")

        geo_lt["type"] = "lt"
        geo_ut["type"] = "ut"

        df1 = pd.concat([geo_lt, bio_lt], axis="columns")
        df2 = pd.concat([geo_ut, bio_ut], axis="columns")

        return pd.concat([df1, df2]).set_index("seed").sort_index()
    except Exception as e:
        raise RuntimeError(f'failed to read stripes for chromosome "{chrom}": {e}')


def _stripes_to_bedpe(
    df: pd.DataFrame, chrom: str, size: int, resolution: int, transpose_policy: Union[str, None]
) -> pd.DataFrame:
    num_stripes = len(df)

    start1_pos = df["left_bound"] * resolution
    end1_pos = np.minimum(df["right_bound"] * resolution, size)
    start2_pos = np.minimum(df["top_bound"] * resolution, size)
    end2_pos = np.minimum(df["bottom_bound"] * resolution, size)

    if transpose_policy is not None:
        if transpose_policy == "transpose_to_ut":
            swap = df["type"] == "lt"
        elif transpose_policy == "transpose_to_lt":
            swap = df["type"] == "ut"
        else:
            raise NotImplementedError
    else:
        swap = None

    if swap is not None:
        start1_pos[swap], start2_pos[swap] = start2_pos[swap], start1_pos[swap]
        end1_pos[swap], end2_pos[swap] = end2_pos[swap], end1_pos[swap]

    return pd.DataFrame(
        {
            "chrom1": [chrom] * num_stripes,
            "start1": start1_pos,
            "end1": end1_pos,
            "chrom2": [chrom] * num_stripes,
            "start2": start2_pos,
            "end2": end2_pos,
        }
    )


def _dump_stripes(f: ResultFile, chrom: str, size: int, resolution: int, cutoff: float, transpose_policy: str):
    df = _read_stripes(f, chrom)
    df = df[df["rel_change"] >= cutoff]
    df = _stripes_to_bedpe(df, chrom, size, resolution, transpose_policy)

    df.to_csv(sys.stdout, sep="\t", index=False, header=False)


def run(
    h5_file: pathlib.Path,
    relative_change_threshold: float,
    transform: Union[str, None],
):
    with ResultFile(h5_file) as f:
        for chrom, size in f.chromosomes.items():
            _dump_stripes(f, chrom, size, f.resolution, relative_change_threshold, transform)



# Copyright (C) 2024 Andrea Raffo <andrea.raffo@ibv.uio.no>
#
# SPDX-License-Identifier: MIT

import datetime
import functools
import json
import pathlib
import shutil
from importlib.metadata import version
from typing import Any, Dict, List, Optional, Sequence, Union

import h5py
import hictkpy
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
import seaborn as sns
from matplotlib.ticker import EngFormatter, ScalarFormatter

from stripepy.utils.stripe import Stripe

no_frills_in_images = False  # TODO safe removal of no_frills_in_images

fruit_punch = sns.blend_palette(["white", "red"], as_cmap=True)


class Result(object):
    """
    A class used to represent the results generated by stripepy call.

    Attributes
    ----------
    chrom: str
        name of the chromosomes to which the Result instance belongs to
    empty: bool
        check whether any stripe has been registered with the Result instance
    roi: Optional[Dict[str, List[int]]]
        the region of interest associated with the Result instance
    min_persistence: float
        the minimum persistence used during computation
    """

    def __init__(self, chrom: str):
        """
        Parameters
        ----------

        chrom: str
            chromosome name
        """
        self._chrom = chrom
        self._roi = None
        self._min_persistence = None

        self._ut_all_minimum_points = None
        self._ut_all_maximum_points = None
        self._ut_persistence_of_all_minimum_points = None
        self._ut_persistence_of_all_maximum_points = None

        self._lt_all_minimum_points = None
        self._lt_all_maximum_points = None
        self._lt_persistence_of_all_minimum_points = None
        self._lt_persistence_of_all_maximum_points = None

        self._ut_persistent_minimum_points = None
        self._ut_persistent_maximum_points = None
        self._ut_persistence_of_minimum_points = None
        self._ut_persistence_of_maximum_points = None
        self._ut_pseudodistribution = None

        self._lt_persistent_minimum_points = None
        self._lt_persistent_maximum_points = None
        self._lt_persistence_of_minimum_points = None
        self._lt_persistence_of_maximum_points = None
        self._lt_pseudodistribution = None

        self._ut_stripes = None
        self._lt_stripes = None

    @property
    def _valid_attributes(self) -> List[str]:
        """
        Get the list of valid attributes
        """
        return [a.removeprefix("_lt_") for a in dir(self) if a.startswith("_lt_")]

    @property
    def empty(self) -> bool:
        return self._lt_stripes is None and self._ut_stripes is None

    @property
    def chrom(self) -> str:
        return self._chrom

    @property
    def roi(self) -> Optional[Dict[str, List[int]]]:
        return self._roi

    @property
    def min_persistence(self) -> float:
        if self._min_persistence is None:
            raise RuntimeError('Attribute "min_persistence" is not set')

        return self._min_persistence

    def get(self, name: str, location: str) -> Union[List[Stripe], npt.NDArray[int], npt.NDArray[float]]:
        """
        Get the value associated with the given attribute name and location.

        Parameters
        ----------
        name: str
            name of the attribute to be fetched
        location: str
            location of the attribute to be fetched. Should be "LT" or "UT"

        Returns
        -------
        attribute
            the value associated with the given name and location.
        """
        if location not in {"LT", "UT"}:
            raise ValueError("Location should be UT or LT")

        attr_name = f"_{location.lower()}_{name}"
        if not hasattr(self, attr_name):
            raise AttributeError(
                f"No attribute named \"{name}\". Valid attributes are: {', '.join(self._valid_attributes)}"
            )

        attr = getattr(self, attr_name)
        if name == "stripes" and attr is None:
            return []

        if attr is None:
            raise RuntimeError(f'Attribute "{name}" for "{location}" is not set')

        return attr

    def get_stripes_descriptor(self, descriptor: str, location: str) -> Union[npt.NDArray[float], npt.NDArray[int]]:
        """
        Get the stripe descriptor for the given location.

        Parameters
        ----------
        descriptor: str
            name of the descriptor to be fetched
        location: str
            location of the attribute to be fetched. Should be "LT" or "UT"

        Returns
        -------
        descriptor
            the value associated with the given descriptor and location.
        """
        if location not in {"LT", "UT"}:
            raise ValueError("Location should be UT or LT")

        if not hasattr(Stripe, descriptor):
            raise AttributeError(f'Stripe instance does not have an attribute named "{descriptor}"')

        stripes = self.get("stripes", location)

        if descriptor in {"seed", "left_bound", "right_bound", "top_bound", "bottom_bound"}:
            dtype = int
        else:
            dtype = float

        return np.array([getattr(stripe, descriptor) for stripe in stripes], dtype=dtype)

    def get_stripe_geo_descriptors(self, location: str) -> pd.DataFrame:
        """
        Fetch all geometric descriptors at once.

        Parameters
        ----------
        location: str
            location of the attribute to be fetched. Should be "LT" or "UT"

        Returns
        -------
        descriptors
            the table with the geometric descriptors associated with the Result instance
        """
        descriptors = [
            "seed",
            "top_persistence",
            "left_bound",
            "right_bound",
            "top_bound",
            "bottom_bound",
        ]

        return pd.DataFrame(
            {descriptor: self.get_stripes_descriptor(descriptor, location) for descriptor in descriptors}
        )

    def get_stripe_bio_descriptors(self, location: str) -> pd.DataFrame:
        """
        Fetch all biological descriptors at once.

        Parameters
        ----------
        location: str
            location of the attribute to be fetched. Should be "LT" or "UT"

        Returns
        -------
        descriptors
            the table with the biological descriptors associated with the Result instance
        """
        descriptors = [
            "inner_mean",
            "outer_mean",
            "rel_change",
            "inner_std",
        ]

        return pd.DataFrame(
            {descriptor: self.get_stripes_descriptor(descriptor, location) for descriptor in descriptors}
        )

    def set_roi(self, coords: Dict[str, List[int]]):
        """
        Set the region of interest (RoI) for the current Result instance.

        Parameters
        ----------
        coords: Dict[str, List[int]]
            a dictionary with the coordinates of the region of interest.
            The dictionary should contain two keys: "genomic" and "matrix".
            The value associated with the "genomic" key should be a list of 4 integers
            representing the genomic coordinates of the region of interest.
            The value associated with the "matrix" key should be a list of 4 integers
            representing the matrix coordinates of the region of interest.
        """
        if self._roi is not None:
            raise RuntimeError("roi has already been set")

        self._roi = coords

    def set_min_persistence(self, min_persistence: float):
        """
        Set the minimum persistence used during computation.

        Parameters
        ----------
        min_persistence: float
        """
        if self._min_persistence is not None:
            raise RuntimeError("min_persistence has already been set")

        self._min_persistence = min_persistence

    def set(self, name: str, data: Union[Sequence[int], Sequence[float], Sequence[Stripe]], location: str):
        """
        Set the attribute corresponding to the given attribute name and location.

        Parameters
        ----------
        name: str
           name of the attribute to be set.
           Supported attributes are:
              * all_minimum_points
              * all_maximum_points
              * persistence_of_all_minimum_points
              * persistence_of_all_maximum_points
              * persistent_minimum_points
              * persistent_maximum_points
              * persistence_of_minimum_points
              * persistence_of_maximum_points
              * pseudodistribution
              * stripes

        data:
            data to be registered with the Result instance
        location: str
            location of the attribute to be registered. Should be "LT" or "UT"
        """
        if location not in {"LT", "UT"}:
            raise ValueError("Location should be UT or LT")

        attr_name = f"_{location.lower()}_{name}"
        if not hasattr(self, attr_name):
            raise AttributeError(
                f"No attribute named \"{name}\". Valid attributes are: {', '.join(self._valid_attributes)}"
            )

        if getattr(self, attr_name) is not None:
            raise RuntimeError(f'Attribute "{name}" for {location} has already been set')

        setattr(self, attr_name, np.array(data))


class ResultFile(object):
    """
    A class used to read and write StripePy results to a HDF5 file.

    Attributes
    ----------
    path: pathlib.Path
        the path to the opened file
    assembly: str
        the name of the reference genome assembly used to generate the file
    resolution: int
        the resolution of the Hi-C matrix used to generate the file
    creation_date: datetime.datetime
        the file creation date
    format: str
        the file format string
    format_url: str
        the URL where the file format is documented
    format_version: int
        the format version of the file currently opened
    generated_by: str
        the name of the tool used to generate the opened file
    normalization: Optional[str]
        the name of the normalization used to generate the data stored in the given file
    chromosomes: Dict[str, int]
        the chromosomes associated with the opened file

    """

    def __init__(self, path: pathlib.Path, mode: str = "r"):
        if mode not in ["r", "w"]:
            raise ValueError('mode should be "r" or "w"')

        self._path = path
        self._mode = mode
        self._chroms = None

        self._h5 = h5py.File(self._path, self._mode)

        if self._mode == "r":
            self._validate(self._h5)
            self._chroms = {
                chrom.decode("utf-8"): size for chrom, size in zip(self._h5["/chroms/name"], self._h5["/chroms/length"])
            }

        self._version = self._h5.attrs.get("format-version", 1)
        self._attrs = dict(self._h5.attrs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._h5.close()

    @property
    def path(self) -> pathlib.Path:
        return self._path

    @functools.cached_property
    def assembly(self) -> str:
        return self._h5.attrs["assembly"]

    @functools.cached_property
    def resolution(self) -> int:
        return int(self._h5.attrs["bin-size"])

    @functools.cached_property
    def creation_date(self) -> datetime.datetime:
        return datetime.datetime.fromisoformat(self._h5.attrs["creation-date"])

    @functools.cached_property
    def format(self) -> str:
        return self._h5.attrs["format"]

    @functools.cached_property
    def format_url(self) -> str:
        return self._h5.attrs["format-url"]

    @functools.cached_property
    def format_version(self) -> int:
        return self._h5.attrs["format-version"]

    @functools.cached_property
    def generated_by(self) -> str:
        return self._h5.attrs["generated-by"]

    @functools.cached_property
    def metadata(self) -> Dict[str, Any]:
        return json.loads(self._h5.attrs["metadata"])

    @functools.cached_property
    def normalization(self) -> Optional[str]:
        norm = self._h5.attrs["normalization"]
        if norm == "NONE":
            return None
        return norm

    @property
    def chromosomes(self) -> Dict[str, int]:
        return self._chroms

    @staticmethod
    def _validate(h5: h5py.File):
        """
        Perform a basic sanity check on the metadata of the current file
        """
        format = h5.attrs.get("format")  # noqa
        format_version = h5.attrs.get("format-version")
        try:
            if format is None:
                raise RuntimeError('attribute "format" is missing')

            if format_version is None:
                raise RuntimeError('attribute "format-version" is missing')

            if format != "HDF5::StripePy":
                raise RuntimeError(f'unrecognized file format: expected "HDF5::StripePy", found "{format}"')

            if format_version != 1:
                raise RuntimeError(
                    f'unsupported file format version "{format_version}". At present only version 1 is supported'
                )
        except RuntimeError as e:
            raise RuntimeError(
                f'failed to validate input file "{h5.filename}": {e}: file is corrupt or was not generated by StripePy.'
            )

    def get_min_persistence(self, chrom: str) -> float:
        """
        Get the minimum persistence associated with the given chromosome.

        Parameters
        ----------
        chrom: str
            chromosome name

        Returns
        -------
            the minimum persistence
        """
        if chrom not in self._chroms:
            raise KeyError(f'File "{self.path}" does not have data for chromosome "{chrom}"')

        return float(self._h5[f"/{chrom}/global-pseudo-distributions"].attrs["min_persistence_used"])

    def get(self, chrom: str, field: str, location: str) -> pd.DataFrame:
        """
        Get the data associated with the given chromosome, field, and location.

        Parameters
        ----------
        chrom: str
            chromosome name
        field: str
            name of the field to be fetched.
            Supported names:
                * pseudodistribution
                * all_minimum_points
                * persistence_of_all_minimum_points
                * all_maximum_points
                * persistence_of_all_maximum_points
                * geo_descriptors
                * bio_descriptors
        location: str
            location of the attribute to be registered. Should be "LT" or "UT"

        Returns
        -------
            the data associated with the given chromosome, field, and location
        """
        if chrom not in self._chroms:
            raise KeyError(f'File "{self.path}" does not have data for chromosome "{chrom}"')

        if location not in {"LT", "UT"}:
            raise ValueError("Location should be UT or LT")

        mappings = {
            "pseudodistribution": f"/{chrom}/global-pseudo-distributions/{location}/pseudo-distribution",
            "all_minimum_points": f"/{chrom}/global-pseudo-distributions/{location}/minima_pts_and_persistence",
            "persistence_of_all_minimum_points": f"/{chrom}/global-pseudo-distributions/{location}/minima_pts_and_persistence",
            "all_maximum_points": f"/{chrom}/global-pseudo-distributions/{location}/maxima_pts_and_persistence",
            "persistence_of_all_maximum_points": f"/{chrom}/global-pseudo-distributions/{location}/maxima_pts_and_persistence",
            "geo_descriptors": f"/{chrom}/stripes/{location}/geo-descriptors",
            "bio_descriptors": f"/{chrom}/stripes/{location}/bio-descriptors",
        }

        if field not in mappings:
            raise KeyError(f"Unknown field \"{field}\". Valid fields are {', '.join(mappings.keys())}")

        path = mappings[field]

        if field not in {"geo_descriptors", "bio_descriptors"}:
            data = self._h5[path][:]
            if field.startswith("persistence"):
                data = data[1, :]
            elif field.endswith("points"):
                data = data[0, :]
            return pd.DataFrame({field: data})

        df = pd.DataFrame(data=self._h5[path], columns=self._h5[path].attrs["col_names"])

        if field == "geo_descriptors":
            df = df.rename(
                columns={
                    "seed persistence": "top_persistence",
                    "L-boundary": "left_bound",
                    "R_boundary": "right_bound",
                    "U-boundary": "top_bound",
                    "D-boundary": "bottom_bound",
                }
            )
            for col in ("seed", "left_bound", "right_bound", "top_bound", "bottom_bound"):
                df[col] = df[col].astype(int)
            return df

        return df.rename(
            columns={
                "inner mean": "inner_mean",
                "outer mean": "outer_mean",
                "relative change": "rel_change",
                "standard deviation": "inner_std",
            }
        )

    def init_file(self, matrix_file: hictkpy.File, normalization: Optional[str], metadata: Dict[str, Any]):
        """
        Initialize the current file.

        This method must be called when opening a ResultFile for writing before adding data to the file.

        Parameters
        ----------
        matrix_file: hictkpy.File
            handle of the file that is used to compute the results stored in this file
        normalization: Optional[str]
            name of the normalization used to compute the results stored in this file
        metadata: Dict[str, Any]
            dictionary with the metadata to be associated with this file.
            The dictionary should contain values that can be encoded as a JSON string
        """
        if normalization is None:
            normalization = "NONE"

        self._chroms = matrix_file.chromosomes(include_ALL=False)

        self._h5.attrs["assembly"] = matrix_file.attributes().get("assembly", "unknown")
        self._h5.attrs["bin-size"] = matrix_file.resolution()
        self._h5.attrs["creation-date"] = datetime.datetime.now().isoformat()
        self._h5.attrs["format"] = "HDF5::StripePy"
        self._h5.attrs["format-url"] = "https://github.com/paulsengroup/StripePy"
        self._h5.attrs["format-version"] = 1
        self._h5.attrs["generated-by"] = f"StripePy v{version('stripepy-hic')}"
        self._h5.attrs["metadata"] = json.dumps(metadata, indent=2)
        self._h5.attrs["normalization"] = normalization

        chroms = matrix_file.chromosomes(include_ALL=False)
        self._h5.create_group("/chroms")
        self._h5.create_dataset("/chroms/name", data=list(chroms.keys()))
        self._h5.create_dataset("/chroms/length", data=list(chroms.values()))

    def write_descriptors(self, result: Result):
        """
        Read the descriptors from the given Result object and write them to the opened file.

        Parameters
        ----------
        result: Result
            results to be added to the opened file
        """
        grp = self._h5.create_group(f"/{result.chrom}/global-pseudo-distributions/")

        grp.attrs.create("min_persistence_used", result.min_persistence)

        for location in ["UT", "LT"]:
            grp = self._h5.create_group(f"/{result.chrom}/global-pseudo-distributions/{location}")
            grp.create_dataset(
                "pseudo-distribution",
                data=result.get("pseudodistribution", location),
                compression="gzip",
                compression_opts=4,
                shuffle=True,
            )
            grp.create_dataset(
                "minima_pts_and_persistence",
                data=np.array(
                    [
                        result.get("all_minimum_points", location),
                        result.get("persistence_of_all_minimum_points", location),
                    ]
                ),
                compression="gzip",
                compression_opts=4,
                shuffle=True,
            )
            grp.create_dataset(
                "maxima_pts_and_persistence",
                data=np.array(
                    [
                        result.get("all_maximum_points", location),
                        result.get("persistence_of_all_maximum_points", location),
                    ]
                ),
                compression="gzip",
                compression_opts=4,
                shuffle=True,
            )

            grp = self._h5.create_group(f"/{result.chrom}/stripes/{location}")
            descriptors = result.get_stripe_geo_descriptors(location)
            dset = grp.create_dataset(
                "geo-descriptors",
                data=descriptors.to_numpy(),
                compression="gzip",
                compression_opts=4,
                shuffle=True,
            )
            dset.attrs.create("col_names", data=descriptors.columns.tolist())

            descriptors = result.get_stripe_bio_descriptors(location)
            dset = grp.create_dataset(
                "bio-descriptors",
                data=descriptors.to_numpy(),
                compression="gzip",
                compression_opts=4,
                shuffle=True,
            )
            dset.attrs.create("col_names", data=descriptors.columns.tolist())


class ANSI:
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    ENDC = "\033[0m"


def remove_and_create_folder(path: pathlib.Path, force: bool):
    path = pathlib.Path(path)

    # Deleting folders:
    if path.exists():
        if not force:
            raise RuntimeError(f"output folder {path} already exists. Pass --force to overwrite it.")
        else:
            shutil.rmtree(path)

    # Create the folder:
    path.mkdir(parents=True)


def create_folders_for_plots(path: pathlib.Path):
    path = pathlib.Path(path)

    folders4plots = [
        path,
        path / "1_preprocessing",
        path / "2_TDA",
        path / "3_shape_analysis",
        path / "4_biological_analysis",
        path / "3_shape_analysis" / "local_pseudodistributions",
    ]

    # Creating folders:
    for folder2create in folders4plots:
        folder2create.mkdir(parents=True)

    return folders4plots


def format_ticks(ax, x=True, y=True, rotate=True):
    """
    Function taken from https://cooltools.readthedocs.io/en/latest/notebooks/viz.html
    :param ax:      an Axes object.
    :param x:       if True, it formats labels in engineering notation for the x-axis
    :param y:       if True, it formats labels in engineering notation for the y-axis
    :param y:       if True, it formats labels in engineering notation for the y-axis
    :param rotate:  if True, it rotates labels in the x-axis by 45 degrees.
    :return:        -
    """

    if x:
        ax.xaxis.set_major_formatter(EngFormatter("b"))
        ax.xaxis.tick_bottom()
    else:
        ax.xaxis.set_major_formatter(ScalarFormatter())
        ax.xaxis.tick_bottom()

    if y:
        ax.yaxis.set_major_formatter(EngFormatter("b"))
    else:
        ax.yaxis.set_major_formatter(ScalarFormatter())

    if rotate:
        ax.tick_params(axis="x", rotation=45)


def pseudodistrib(
    pseudo_distrib,
    IoI,
    resolution,
    coords2scatter=None,
    colors=None,
    output_folder=None,
    file_name=None,
    title=None,
    display=False,
):
    # TODO remove
    """
    :param pseudo_distrib:          1D ndarray representing a scalar function sampled over a uniform mesh
    :param  IoI:                    refers to the Interval of Interest [IoI[0], IoI[1]] (e.g., in genomic coordinates)
                                    where the scalar function was sampled on; see also plot_in_bp
    :param resolution:              resolution of the Hi-C matrix
    :param coords2scatter:          list of lists of genomic coordinates; each list of genomic coordinates determines
                                    a point cloud as follows: for genomic coordinate x, we sample the value of
                                    pseudo_distrib at x; each point cloud is scatterplotted with a specific color,
                                    potentially given in input, read below;
                                    if set to None, nothing happens
    :param colors:                  list of colors, one color per list of genomic coordinates (see coords2scatter);
                                    if set to None, use red
    :param output_folder:           path to folder where to save the image
    :param file_name:               name of the file to be created
    :param title:                   title to give to the image
    :param display:                 if False, it does not display the plot
    :return:                        -
    """

    fig, ax = plt.subplots(1, 1)
    ax.plot(
        range(IoI[0], IoI[1], resolution),
        pseudo_distrib[int(IoI[0] / resolution) : int(IoI[1] / resolution)],
        color="red",
        linewidth=0.5,
        linestyle="solid",
    )
    if coords2scatter is not None:
        for n, cur_coords2scatter in enumerate(coords2scatter):
            if colors is not None:
                color = colors[n]
            else:
                color = "red"
            ax.plot(
                [cur_coord2scatter * resolution for cur_coord2scatter in cur_coords2scatter],
                pseudo_distrib[cur_coords2scatter],
                marker=".",
                linestyle="",
                markersize=6 * 1.5,
                color=color,
            )
    ax.xaxis.set_major_formatter(EngFormatter("b"))

    if no_frills_in_images is False:
        if title is not None:
            fig.suptitle(title)

    plt.ylim((0, 1))
    ax.set_xlabel("genomic coordinates (bp)")
    ax.set_ylabel("pseudo-distribution")
    fig.tight_layout()
    ax.grid(True)
    # plt.axis('scaled')

    if output_folder is not None and file_name is not None:
        plt.savefig(f"{output_folder}/{file_name}", bbox_inches="tight")

    if display:
        plt.show()
    plt.close()


def pseudodistrib_and_HIoIs(
    pseudo_distrib, IoIs, resolution, colors=None, output_folder=None, file_name=None, title=None, display=False
):
    # TODO remove
    """
    :param pseudo_distrib:          1D ndarray representing a scalar function sampled over a uniform mesh
    :param  IoIs:                   list of lists, where the innermost lists are pairs of coordinates; the first pair
                                    refers to the Interval of Interest [IoI[0], IoI[1]] (e.g., in genomic coordinates)
                                    where pseudo-distribution is plotted; the remaining pairs define sub-regions to be
                                    plotted in (potentially) different colors
                                    where the scalar function was sampled on; see also plot_in_bp
    :param resolution:              resolution of the Hi-C matrix
    :param colors:                  list of colors, one color per pair of genomic coordinates (see IoIs);
                                    if set to None, use red
    :param output_folder:           path to folder where to save the image
    :param file_name:               name of the file to be created
    :param title:                   title to give to the image
    :param display:                 if False, it does not display the plot
    :return:                        -
    """
    fig, ax = plt.subplots(1, 1)
    for IoI, color in zip(IoIs, colors):
        ax.plot(
            range(IoI[0], IoI[1], resolution),
            pseudo_distrib[int(IoI[0] / resolution) : int(IoI[1] / resolution)],
            color=color,
            linewidth=0.5,
            linestyle="solid",
        )
    ax.xaxis.set_major_formatter(EngFormatter("b"))
    if no_frills_in_images is False:
        if title is not None:
            fig.suptitle(title)
    ax.set_xlabel("genomic coordinates (bp)")
    ax.set_ylabel("pseudo-distribution")
    fig.tight_layout()
    ax.grid(True)
    # plt.axis('scaled')
    if output_folder is not None and file_name is not None:
        plt.savefig(f"{output_folder}/{file_name}", bbox_inches="tight")
    if display:
        plt.show()
    plt.close()


def HiC_and_sites(
    I,
    sites,
    RoI,
    resolution,
    where=None,
    plot_in_bp=False,
    output_folder=None,
    file_name=None,
    title=None,
    display=False,
):  # TODO remove
    """
    :param I:                  Hi-C matrix to be plotted as image and saved
    :param sites:              list of locations of interest
    :param  RoI:               refers to the region of interest [RoI[0], RoI[1]]x[RoI[2], RoI[3]]
                               (e.g., in genomic coordinates); see also plot_in_bp
    :param where:              if "lower" (resp. "upper"), then it plots sites only on the lower (resp. upper) part of
                               the Hi-C matrix; otherwise, it plots sites spanning the whole Hi-C matrix
    :param plot_in_bp:         if True, labels are set in base pairs
    :param output_folder:      path to folder where to save the image
    :param file_name:          name of the file to be created
    :param title:              title to give to the image
    :param display:            if False, it does not display the plot
    :return:                   -
    """

    fig, ax = plt.subplots(1, 1)
    img = ax.matshow(I, vmax=np.amax(I), extent=(RoI[0], RoI[1], RoI[3], RoI[2]), cmap=fruit_punch)
    if plot_in_bp:
        format_ticks(ax)

    for site in [site * resolution for site in sites]:
        if where == "lower":
            ax.plot(
                [site, site], [site, RoI[1] - 1 * resolution], color=(0.0, 0.0, 1.0), linestyle="dashed", linewidth=1
            )
        elif where == "upper":
            ax.plot(
                [site, site], [RoI[0] + 1 * resolution, site], color=(0.0, 0.0, 1.0), linestyle="dashed", linewidth=1
            )
        else:
            ax.plot(
                [site, site],
                [RoI[0] + 1 * resolution, RoI[2] - 1 * resolution],
                color=(0.0, 0.0, 1.0),
                linestyle="dashed",
                linewidth=1,
            )

    if no_frills_in_images is False:
        plt.colorbar(img)
        if title is not None:
            fig.suptitle(title)
    else:
        plt.axis("off")

    fig.set_dpi(256)
    plt.axis("scaled")
    fig.tight_layout()

    if output_folder is not None and file_name is not None:
        plt.savefig(f"{output_folder}/{file_name}", bbox_inches="tight")

    if display is True:
        plt.show()
    plt.close()


def HiC_and_HIoIs(
    I,
    HIoIs,
    RoI,
    resolution,
    where=None,
    plot_in_bp=False,
    output_folder=None,
    file_name=None,
    title=None,
    display=False,
):
    # TODO: remove
    """
    :param I:                   Hi-C matrix to be plotted as image and saved
    :param HIoIs:               list of lists, where the innermost lists are pairs of elements
    :param  RoI:                refers to the region of interest [RoI[0], RoI[1]]x[RoI[2], RoI[3]]
                                (e.g., in genomic coordinates); see also plot_in_bp
    :param resolution:              resolution of the Hi-C matrix
    :param where:               if "lower" (resp. "upper"), then it plots sites only on the lower (resp. upper) part of
                                the Hi-C matrix; otherwise, it plots sites spanning the whole Hi-C matrix
    :param plot_in_bp:          if True, labels are set in base pairs
    :param output_folder:       path to folder where to save the image
    :param file_name:           name of the file to be created
    :param title:               title to give to the image
    :param display:             if False, it does not display the plot
    :return:                    -
    """

    fig, ax = plt.subplots(1, 1)
    img = ax.matshow(I, vmax=np.amax(I), extent=(RoI[0], RoI[1], RoI[3], RoI[2]), cmap=fruit_punch)
    if plot_in_bp:
        format_ticks(ax)

    for HIoI in HIoIs:
        if where == "lower":
            ax.plot(
                [RoI[0] + HIoI[0], RoI[0] + HIoI[0]],
                [RoI[2] + HIoI[0], RoI[2] + (I.shape[0] - 1) * resolution],
                color=(0.0, 0.0, 1.0),
                linestyle="dashed",
                linewidth=1,
            )
            ax.plot(
                [RoI[0] + HIoI[1] + resolution, RoI[0] + HIoI[1] + resolution],
                [RoI[2] + HIoI[1], RoI[2] + (I.shape[0] - 1) * resolution],
                color=(0.0, 0.0, 1.0),
                linestyle="dashed",
                linewidth=1,
            )
        elif where == "upper":
            ax.plot(
                [RoI[0] + HIoI[0], RoI[0] + HIoI[0]],
                [RoI[2] + 1 * resolution, RoI[2] + HIoI[0]],
                color=(0.0, 0.0, 1.0),
                linestyle="dashed",
                linewidth=1,
            )
            ax.plot(
                [RoI[0] + HIoI[1] + resolution, RoI[0] + HIoI[1] + resolution],
                [RoI[2] + 1 * resolution, RoI[2] + HIoI[1] + resolution],
                color=(0.0, 0.0, 1.0),
                linestyle="dashed",
                linewidth=1,
            )
        else:
            ax.plot(
                [RoI[0] + HIoI[0], RoI[0] + HIoI[0]],
                [RoI[2] + 1 * resolution, RoI[2] + (I.shape[0] - 1) * resolution],
                color=(0.0, 0.0, 1.0),
                linestyle="dashed",
                linewidth=1,
            )
            ax.plot(
                [RoI[0] + HIoI[1] + resolution, RoI[0] + HIoI[1] + resolution],
                [RoI[2] + 1 * resolution, RoI[2] + (I.shape[0] - 1) * resolution],
                color=(0.0, 0.0, 1.0),
                linestyle="dashed",
                linewidth=1,
            )

    if no_frills_in_images is False:
        plt.colorbar(img)
        if title is not None:
            fig.suptitle(title)
    else:
        plt.axis("off")

    fig.set_dpi(256)
    plt.axis("scaled")
    fig.tight_layout()

    if output_folder is not None and file_name is not None:
        plt.savefig(f"{output_folder}/{file_name}", bbox_inches="tight")

    if display:
        plt.show()
    plt.close()


def plot_stripes(
    I,
    LT_HIoIs,
    LT_VIoIs,
    UT_HIoIs,
    UT_VIoIs,
    RoI,
    resolution,
    plot_in_bp=False,
    output_folder=None,
    file_name=None,
    title=None,
    display=False,
):
    # TODO remove
    """
    :param I:                  Hi-C matrix to be plotted as image and saved
    :param LT_HIoIs:           Horizontal Intervals of Interest (lower-triangular part)
    :param LT_VIoIs:           Vertical Intervals of Interest (lower-triangular part)
    :param UT_HIoIs:           Horizontal Intervals of Interest (upper-triangular part)
    :param UT_VIoIs:           Vertical Intervals of Interest (upper-triangular part)
    :param RoI:                refers to the region of interest [RoI[0], RoI[1]]x[RoI[2], RoI[3]]
                               (e.g., in genomic coordinates); see also plot_in_bp
    :param plot_in_bp:         if True, labels are set in base pairs
    :param output_folder:      path to folder where to save the image
    :param file_name:          name of the file to be created
    :param title:              title to give to the image
    :param display:            if False, it does not display the plot
    :return:                   -
    """

    fig, ax = plt.subplots(1, 1)
    ax.matshow(I, vmax=np.amax(I), extent=(RoI[0], RoI[1], RoI[3], RoI[2]), cmap=fruit_punch)
    if plot_in_bp:
        format_ticks(ax)

    # Upper-triangular candidates:
    for num_window, (HIoI, VIoI) in enumerate(zip(UT_HIoIs, UT_VIoIs)):
        low_idx_up_opt = max(1, VIoI[0]) * resolution
        low_idx_dw_opt = min(VIoI[1], I.shape[0] - 1) * resolution
        low_idx_sx_opt = max(1, HIoI[0]) * resolution
        low_idx_dx_opt = min(HIoI[1], I.shape[1] - 1) * resolution
        if low_idx_up_opt != low_idx_dw_opt and low_idx_sx_opt != low_idx_dx_opt:
            ax.plot(
                [RoI[0] + low_idx_sx_opt, RoI[2] + low_idx_sx_opt],
                [RoI[0] + low_idx_dw_opt, RoI[2] + low_idx_up_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )
            ax.plot(
                [RoI[0] + low_idx_dx_opt, RoI[2] + low_idx_dx_opt],
                [RoI[0] + low_idx_dw_opt, RoI[2] + low_idx_up_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )
            ax.plot(
                [RoI[0] + low_idx_sx_opt, RoI[2] + low_idx_dx_opt],
                [RoI[0] + low_idx_dw_opt, RoI[2] + low_idx_dw_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )
            ax.plot(
                [RoI[0] + low_idx_sx_opt, RoI[2] + low_idx_dx_opt],
                [RoI[0] + low_idx_up_opt, RoI[2] + low_idx_up_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )
            # plt.text(RoI[0] + low_idx_sx_opt, RoI[2] + low_idx_dw_opt + 15, str(num_window))

    for num_window, (HIoI, VIoI) in enumerate(zip(LT_HIoIs, LT_VIoIs)):
        upp_idx_up_opt = max(1, VIoI[0]) * resolution
        upp_idx_dw_opt = min(VIoI[1], I.shape[0] - 1) * resolution
        upp_idx_sx_opt = max(1, HIoI[0]) * resolution
        upp_idx_dx_opt = min(HIoI[1], I.shape[0] - 1) * resolution
        if upp_idx_up_opt != upp_idx_dw_opt and upp_idx_sx_opt != upp_idx_dx_opt:
            ax.plot(
                [RoI[0] + upp_idx_sx_opt, RoI[2] + upp_idx_sx_opt],
                [RoI[0] + upp_idx_dw_opt, RoI[2] + upp_idx_up_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )
            ax.plot(
                [RoI[0] + upp_idx_dx_opt, RoI[2] + upp_idx_dx_opt],
                [RoI[0] + upp_idx_dw_opt, RoI[2] + upp_idx_up_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )
            ax.plot(
                [RoI[0] + upp_idx_sx_opt, RoI[2] + upp_idx_dx_opt],
                [RoI[0] + upp_idx_dw_opt, RoI[2] + upp_idx_dw_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )
            ax.plot(
                [RoI[0] + upp_idx_sx_opt, RoI[2] + upp_idx_dx_opt],
                [RoI[0] + upp_idx_up_opt, RoI[2] + upp_idx_up_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )
            # plt.text(RoI[0] + upp_idx_sx_opt, RoI[2] + upp_idx_up_opt - 15, str(num_window))

    if no_frills_in_images is False:
        if title is not None:
            fig.suptitle(title)
    else:
        plt.axis("off")

    fig.set_dpi(256)
    fig.tight_layout()

    if output_folder is not None and file_name is not None:
        plt.savefig(f"{output_folder}/{file_name}", bbox_inches="tight")

    if display:
        plt.show()
    plt.close()


def plot_stripes_and_peaks(
    I,
    LT_HIoIs,
    LT_VIoIs,
    UT_HIoIs,
    UT_VIoIs,
    LT_peaks_ids,
    UT_peaks_ids,
    RoI,
    resolution,
    plot_in_bp=False,
    output_folder=None,
    file_name=None,
    title=None,
    display=False,
):
    # TODO CORREGGERE LA TRASLAZIONE DI ROI
    """
    :param I:                  Hi-C matrix to be plotted as image and saved
    :param LT_HIoIs:           Horizontal Intervals of Interest (lower-triangular part)
    :param LT_VIoIs:           Vertical Intervals of Interest (lower-triangular part)
    :param UT_HIoIs:           Horizontal Intervals of Interest (upper-triangular part)
    :param UT_VIoIs:           Vertical Intervals of Interest (upper-triangular part)
    :param RoI:                refers to the region of interest [RoI[0], RoI[1]]x[RoI[2], RoI[3]]
                               (e.g., in genomic coordinates); see also plot_in_bp
    :param plot_in_bp:         if True, labels are set in base pairs
    :param output_folder:      path to folder where to save the image
    :param file_name:          name of the file to be created
    :param title:              title to give to the image
    :param display:            if False, it does not display the plot
    :return:                   -
    """

    fig, ax = plt.subplots(1, 1)
    ax.matshow(I, vmax=np.amax(I), extent=(RoI[0], RoI[1], RoI[3], RoI[2]), cmap=fruit_punch)
    if plot_in_bp:
        format_ticks(ax)

    # Upper-triangular candidates:
    for num_window, (HIoI, VIoI, peaks_ids) in enumerate(zip(UT_HIoIs, UT_VIoIs, UT_peaks_ids)):
        low_idx_up_opt = max(1, VIoI[0]) * resolution
        low_idx_dw_opt = min(VIoI[1], I.shape[0] - 1) * resolution
        low_idx_sx_opt = max(1, HIoI[0]) * resolution
        low_idx_dx_opt = min(HIoI[1], I.shape[1] - 1) * resolution
        if low_idx_up_opt != low_idx_dw_opt and low_idx_sx_opt != low_idx_dx_opt:
            ax.plot(
                [RoI[0] + low_idx_sx_opt, RoI[2] + low_idx_sx_opt],
                [RoI[0] + low_idx_dw_opt, RoI[2] + low_idx_up_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )
            ax.plot(
                [RoI[0] + low_idx_dx_opt, RoI[2] + low_idx_dx_opt],
                [RoI[0] + low_idx_dw_opt, RoI[2] + low_idx_up_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )
            ax.plot(
                [RoI[0] + low_idx_sx_opt, RoI[2] + low_idx_dx_opt],
                [RoI[0] + low_idx_dw_opt, RoI[2] + low_idx_dw_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )
            ax.plot(
                [RoI[0] + low_idx_sx_opt, RoI[2] + low_idx_dx_opt],
                [RoI[0] + low_idx_up_opt, RoI[2] + low_idx_up_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )

            transl_peak_roots_ids = [RoI[0] + (low_idx_sx_opt + low_idx_dx_opt) / 2 for _ in peaks_ids]
            transl_peaks_ids = [peak_idx * resolution + RoI[2] for peak_idx in peaks_ids]
            ax.scatter(transl_peak_roots_ids, transl_peaks_ids, c="blue")

    for num_window, (HIoI, VIoI, peaks_ids) in enumerate(zip(LT_HIoIs, LT_VIoIs, LT_peaks_ids)):
        upp_idx_up_opt = max(1, VIoI[0]) * resolution
        upp_idx_dw_opt = min(VIoI[1], I.shape[0] - 1) * resolution
        upp_idx_sx_opt = max(1, HIoI[0]) * resolution
        upp_idx_dx_opt = min(HIoI[1], I.shape[0] - 1) * resolution
        if upp_idx_up_opt != upp_idx_dw_opt and upp_idx_sx_opt != upp_idx_dx_opt:
            ax.plot(
                [RoI[0] + upp_idx_sx_opt, RoI[2] + upp_idx_sx_opt],
                [RoI[0] + upp_idx_dw_opt, RoI[2] + upp_idx_up_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )
            ax.plot(
                [RoI[0] + upp_idx_dx_opt, RoI[2] + upp_idx_dx_opt],
                [RoI[0] + upp_idx_dw_opt, RoI[2] + upp_idx_up_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )
            ax.plot(
                [RoI[0] + upp_idx_sx_opt, RoI[2] + upp_idx_dx_opt],
                [RoI[0] + upp_idx_dw_opt, RoI[2] + upp_idx_dw_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )
            ax.plot(
                [RoI[0] + upp_idx_sx_opt, RoI[2] + upp_idx_dx_opt],
                [RoI[0] + upp_idx_up_opt, RoI[2] + upp_idx_up_opt],
                "blue",
                linestyle="dashed",
                linewidth=1,
            )

            transl_peak_roots_ids = [RoI[0] + (upp_idx_sx_opt + upp_idx_dx_opt) / 2 for _ in peaks_ids]
            transl_peaks_ids = [peak_idx * resolution + RoI[2] for peak_idx in peaks_ids]
            ax.scatter(transl_peak_roots_ids, transl_peaks_ids, c="blue")
            # plt.text(RoI[0] + upp_idx_sx_opt, RoI[2] + upp_idx_up_opt - 15, str(num_window))

    if no_frills_in_images is False:
        if title is not None:
            fig.suptitle(title)
    else:
        plt.axis("off")

    fig.set_dpi(256)
    fig.tight_layout()

    if output_folder is not None and file_name is not None:
        plt.savefig(f"{output_folder}/{file_name}", bbox_inches="tight")

    if display:
        plt.show()
    plt.close()



# Copyright (C) 2024 Andrea Raffo <andrea.raffo@ibv.uio.no>
#
# SPDX-License-Identifier: MIT

import numpy as np

from . import stripe

# TODO All functions defined in this file appear to be unused.
#      Review!


def filter_stripes(candida_stripes, threshold):
    rel_changes = np.array([candida_stripe.rel_change for candida_stripe in candida_stripes])
    keep_candidate = rel_changes > threshold

    return keep_candidate


def filter_L_stripes(I, sites, HIoIs, VIoIs, threshold, output_folder=None):
    """
    :param I:               Hi-C matrix
    :param HIoIs:           Horizontal Intervals of Interest
    :param VIoIs:           Vertical Intervals of Interest
    :param threshold:       threshold to filter out candidates
    :return:                   -
    """

    # Candidate stripes:
    candida_stripes = []

    # Indices stripes to keep or discard:
    candida2keep = []
    candida2disc = []

    # Regions 4 check:
    L_HIoI_regions4check = []
    L_VIoI_regions4check = []

    # Statistical analysis of the avg pixel intensity in a neighborhood:

    inner_means = [0] * len(sites)

    for num_cand, (site, HIoI, VIoI) in enumerate(zip(sites, HIoIs, VIoIs)):

        # Update list of candidate stripes:
        candida_stripe = stripe.Stripe(site, HIoI[0], HIoI[1], VIoI[0], VIoI[1])

        # Restricting I to the rectangular domain defined by current HIoI and VIoI:
        convex_comb = int(round(0.99 * VIoI[0] + 0.01 * VIoI[1]))
        # restrI = I[HIoI[1]:VIoI[1], HIoI[0]:HIoI[1]]
        restrI = I[convex_comb : VIoI[1], HIoI[0] : HIoI[1]]

        # Horizontal enlargements:
        enl_HIoI_1 = max(0, HIoI[0] - 3)
        enl_HIoI_2 = min(I.shape[1], HIoI[1] + 3)

        # Five-number statistics inside the candidate stripe:
        candida_stripe.inner_descriptors["five-number"] = [
            np.min(restrI),
            np.percentile(restrI, 25),
            np.percentile(restrI, 50),
            np.percentile(restrI, 75),
            np.max(restrI),
        ]
        candida_stripe.inner_descriptors["mean"] = np.mean(restrI)
        candida_stripe.inner_descriptors["std"] = np.std(restrI)

        inner_means[num_cand] = candida_stripe.inner_descriptors["mean"]

        # Mean intensity - left neighborhood:
        candida_stripe.outer_descriptors["l-mean"] = np.mean(I[convex_comb : VIoI[1], enl_HIoI_1 : HIoI[0]])

        # Mean intensity - right neighborhood:
        candida_stripe.outer_descriptors["r-mean"] = np.mean(I[convex_comb : VIoI[1], HIoI[1] : enl_HIoI_2])

        # Mean intensity:
        candida_stripe.outer_descriptors["mean"] = (
            candida_stripe.outer_descriptors["l-mean"] + candida_stripe.outer_descriptors["r-mean"]
        ) / 2

        # Relative change in mean intensity:
        candida_stripe.rel_change = (
            (candida_stripe.inner_descriptors["mean"] - candida_stripe.outer_descriptors["mean"])
            / (candida_stripe.outer_descriptors["mean"])
            * 100
        )

        candida_stripes.append(candida_stripe)

        # Filtration:
        if candida_stripe.rel_change < threshold:
            candida2disc += [num_cand]
        else:
            candida2keep += [num_cand]

        L_HIoI_regions4check += [[HIoI[1], enl_HIoI_2]]
        L_VIoI_regions4check += [[convex_comb, VIoI[1]]]

    if output_folder is not None:
        np.savetxt(f"{output_folder}/LT_inner_means_{threshold:.2f}.txt", inner_means, delimiter=",")

    return candida_stripes, candida2keep, candida2disc, L_HIoI_regions4check, L_VIoI_regions4check


def filter_U_stripes(I, sites, HIoIs, VIoIs, threshold, output_folder=None):
    """
    :param I:               Hi-C matrix
    :param HIoIs:           Horizontal Intervals of Interest
    :param VIoIs:           Vertical Intervals of Interest
    :param threshold:       threshold to filter out candidates
    :return:                   -
    """

    # Candidate stripes:
    candida_stripes = []

    # Indices stripes to keep or discard:
    candida2keep = []
    candida2disc = []

    # Regions 4 check:
    U_HIoI_regions4check = []
    U_VIoI_regions4check = []

    inner_means = [0] * len(sites)

    # Statistical analysis of the avg pixel intensity in a neighborhood:
    for num_cand, (site, HIoI, VIoI) in enumerate(zip(sites, HIoIs, VIoIs)):

        # Update list of candidate stripes:
        candida_stripe = stripe.Stripe(HIoI[0], HIoI[1], VIoI[0], VIoI[1], seed=site)

        # Restricting Iproc1 to the rectangular domain defined by current HIoI and VIoI:
        convex_comb = int(round(0.01 * VIoI[0] + 0.99 * VIoI[1]))
        restrI = I[VIoI[0] : convex_comb, HIoI[0] : HIoI[1]]

        # Horizontal enlargements:
        enl_HIoI_1 = max(0, HIoI[0] - 3)
        enl_HIoI_2 = min(I.shape[1], HIoI[1] + 3)

        print(f"{num_cand}) {site} {HIoI} {VIoI} {enl_HIoI_1} {enl_HIoI_2}")

        # Five-number statistics inside the candidate stripe:
        candida_stripe.inner_descriptors["five-number"] = [
            np.min(restrI),
            np.percentile(restrI, 25),
            np.percentile(restrI, 50),
            np.percentile(restrI, 75),
            np.max(restrI),
        ]
        candida_stripe.inner_descriptors["mean"] = np.mean(restrI)
        candida_stripe.inner_descriptors["std"] = np.std(restrI)

        inner_means[num_cand] = candida_stripe.inner_descriptors["mean"]

        # Mean intensity - left neighborhood:
        # l_mean = np.mean(I[HIoI[0]:VIoI[1], enl_HIoI_1:HIoI[0]])
        candida_stripe.outer_descriptors["l-mean"] = np.mean(I[VIoI[0] : convex_comb, enl_HIoI_1 : HIoI[0]])

        # Mean intensity - right neighborhood:
        # r_mean = np.mean(I[enl_HIoI_2:VIoI[1], HIoI[1]:enl_HIoI_2])
        candida_stripe.outer_descriptors["r-mean"] = np.mean(I[VIoI[0] : convex_comb, HIoI[1] : enl_HIoI_2])

        # Mean intensity:
        candida_stripe.outer_descriptors["mean"] = (
            candida_stripe.outer_descriptors["l-mean"] + candida_stripe.outer_descriptors["r-mean"]
        ) / 2

        # Relative change in mean intensity:
        candida_stripe.rel_change = (
            (candida_stripe.inner_descriptors["mean"] - candida_stripe.outer_descriptors["mean"])
            / (candida_stripe.outer_descriptors["mean"])
            * 100
        )

        print(
            f"{num_cand}) "
            f"{candida_stripe.inner_descriptors["mean"]}\t"
            f"{candida_stripe.outer_descriptors["mean"]}\t"
            f"{candida_stripe.outer_descriptors["l-mean"]}\t"
            f"{candida_stripe.outer_descriptors["r-mean"]}\t{candida_stripe.rel_change}"
        )

        candida_stripes.append(candida_stripe)

        # Filtration:
        if candida_stripe.rel_change < threshold:
            candida2disc += [num_cand]
        else:
            candida2keep += [num_cand]

        U_HIoI_regions4check += [[HIoI[1], enl_HIoI_2]]
        U_VIoI_regions4check += [[convex_comb, VIoI[1]]]

    if output_folder is not None:
        np.savetxt(f"{output_folder}/UT_inner_means_{threshold:.2f}.txt", inner_means, delimiter=",")

    return candida_stripes, candida2keep, candida2disc, U_HIoI_regions4check, U_VIoI_regions4check



# Copyright (C) 2024 Roberto Rossini <roberros@uio.no>
#
# SPDX-License-Identifier: MIT

import gc
import pathlib
from typing import Dict, Optional

import hictkpy as htk
import pandas as pd

from .data import generate_chromosomes


def generate_singleres_test_file(
    path: pathlib.Path, resolution: int, chromosomes: Optional[Dict[str, int]] = None
) -> pathlib.Path:
    if chromosomes is None:
        chromosomes = generate_chromosomes()

    writer = htk.cooler.FileWriter(str(path), resolution=resolution, chromosomes=chromosomes)
    writer.add_pixels(pd.DataFrame({"bin1_id": [0], "bin2_id": [0], "count": 1}))
    writer.finalize()

    del writer
    gc.collect()

    return pathlib.Path(path)



# Copyright (C) 2024 Roberto Rossini <roberros@uio.no>
#
# SPDX-License-Identifier: MIT

import functools
from typing import Dict


@functools.cache
def generate_chromosomes() -> Dict[str, int]:
    return {
        "chr2L": 23513712,
        "chr2R": 25286936,
        "chr3L": 28110227,
        "chr3R": 32079331,
        "chr4": 1348131,
        "chrX": 23542271,
        "chrY": 3667352,
        "chrM": 19524,
    }



# Copyright (C) 2024 Roberto Rossini <roberros@uio.no>
#
# SPDX-License-Identifier: MIT

import pathlib
import tempfile
from typing import List

import pytest

from stripepy.others import cmap_loading

from .common.cool import generate_singleres_test_file


def _discretize(v: List[int], factor: int) -> List[int]:
    assert factor > 0
    return [(x + factor - 1) // factor for x in v]


@pytest.mark.unit
class TestCmapLoading:
    def test_invalid_paths(self, tmpdir):
        tmpdir = pathlib.Path(tmpdir)
        with pytest.raises(RuntimeError):
            cmap_loading(tmpdir / "foo", 5000)

    def test_invalid_formats(self, tmpdir):
        invalid_file = pathlib.Path(__file__).resolve()
        folder = pathlib.Path(tmpdir)
        with pytest.raises(RuntimeError):
            cmap_loading(invalid_file, 5000)

        with pytest.raises(RuntimeError):
            cmap_loading(folder, 5000)

    def test_invalid_resolutions(self, tmpdir):
        with tempfile.NamedTemporaryFile(dir=tmpdir) as clr:
            clr.close()
            path_to_clr = generate_singleres_test_file(pathlib.Path(clr.name), 1000)

            with pytest.raises(RuntimeError):
                cmap_loading(path_to_clr, 5000)
            with pytest.raises(TypeError, match="must be an integer"):
                cmap_loading(path_to_clr, 1000.0)
            with pytest.raises(ValueError, match="must be greater than zero"):
                cmap_loading(path_to_clr, -1)
            with pytest.raises(ValueError, match="must be greater than zero"):
                cmap_loading(path_to_clr, 0)

    def test_valid_files(self, tmpdir):
        tmpdir = pathlib.Path(tmpdir)
        chromosomes = {"A": 100_000, "B": 50_000, "C": 10_000}
        resolution = 1000
        path_to_clr = generate_singleres_test_file(tmpdir / "test.cool", resolution, chromosomes)

        f, starts, ends, sizes = cmap_loading(path_to_clr, resolution)
        assert f.resolution() == 1000
        assert len(starts) == len(chromosomes)
        assert len(starts) == len(ends)
        assert len(starts) == len(sizes)

        assert starts == _discretize([0, 100_000, 150_000], resolution)
        assert ends == _discretize([100_000, 150_000, 160_000], resolution)
        assert sizes == list(chromosomes.values())

        with pytest.raises(Exception):
            cmap_loading(path_to_clr, resolution + 1)
